from __future__ import annotations

import hashlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Iterable

import graphene

from infrahub import config
from infrahub.core.attribute import String
from infrahub.core.constants import InfrahubKind, RelationshipCardinality, RelationshipKind
from infrahub.core.schema import (
    AttributeSchema,
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    RelationshipSchema,
    TemplateSchema,
)
from infrahub.graphql.mutations.attribute import BaseAttributeCreate, BaseAttributeUpdate
from infrahub.graphql.mutations.graphql_query import InfrahubGraphQLQueryMutation
from infrahub.graphql.mutations.profile import InfrahubProfileMutation
from infrahub.graphql.types.metadata import OrderInput
from infrahub.types import ATTRIBUTE_TYPES, InfrahubDataType, get_attribute_type

from .constants import NODE_METADATA_TYPE, RELATIONSHIP_METADATA_TYPE
from .directives import DIRECTIVES
from .enums import generate_graphql_enum, get_enum_attribute_type_name
from .metrics import SCHEMA_GENERATE_GRAPHQL_METRICS
from .mutations.action import InfrahubTriggerRuleMatchMutation, InfrahubTriggerRuleMutation
from .mutations.artifact_definition import InfrahubArtifactDefinitionMutation
from .mutations.ipam import (
    InfrahubIPAddressMutation,
    InfrahubIPNamespaceMutation,
    InfrahubIPPrefixMutation,
)
from .mutations.main import InfrahubMutation
from .mutations.menu import InfrahubCoreMenuMutation
from .mutations.proposed_change import InfrahubProposedChangeMutation
from .mutations.repository import InfrahubRepositoryMutation
from .mutations.resource_manager import (
    InfrahubNumberPoolMutation,
)
from .mutations.webhook import InfrahubWebhookMutation
from .registry import registry
from .resolvers.account_metadata import account_metadata_resolver
from .resolvers.ipam import ipam_paginated_list_resolver
from .resolvers.resolver import (
    account_resolver,
    ancestors_resolver,
    default_paginated_list_resolver,
    default_resolver,
    descendants_resolver,
    many_relationship_resolver,
    parent_field_name_resolver,
    single_relationship_resolver,
)
from .schema import InfrahubBaseMutation, InfrahubBaseQuery
from .subscription import InfrahubBaseSubscription
from .types import (
    InfrahubInterface,
    InfrahubObject,
    PaginatedObjectPermission,
    RelatedIPAddressNodeInput,
    RelatedIPPrefixNodeInput,
    RelatedNodeInput,
    Upload,
)
from .types.attribute import BaseAttribute as BaseAttributeType
from .types.attribute import TextAttributeType
from .types.branch import InfrahubBranchEdge
from .types.context import ContextInput
from .types.event import EVENT_TYPES
from .types.node import InfrahubObjectWithoutMeta

if TYPE_CHECKING:
    from graphql import GraphQLSchema

    from infrahub.core.schema.schema_branch import SchemaBranch


class DeleteInput(graphene.InputObjectType):
    id = graphene.String(required=False)
    hfid = graphene.List(of_type=graphene.String, required=False)


GraphQLTypes = type[InfrahubMutation] | type[BaseAttributeType] | type[graphene.Interface] | type[graphene.ObjectType]


@dataclass
class GraphqlMutations:
    create: type[InfrahubMutation]
    update: type[InfrahubMutation]
    upsert: type[InfrahubMutation]
    delete: type[InfrahubMutation]


@dataclass
class InterfaceReference:
    reference: type[graphene.Interface]
    reference_hash: str


@dataclass
class InfrahubObjectReference:
    reference: type[InfrahubObject]
    reference_hash: str


@dataclass
class InfrahubEdgedReference:
    reference: type[InfrahubObject]
    reference_hash: str


def get_attr_kind(node_schema: MainSchemaTypes, attr_schema: AttributeSchema) -> str:
    if not config.SETTINGS.experimental_features.graphql_enums or not attr_schema.enum:
        return attr_schema.kind
    return get_enum_attribute_type_name(node_schema=node_schema, attr_schema=attr_schema)


class GraphQLSchemaManager:
    def __init__(self, schema: SchemaBranch) -> None:
        self.schema = schema
        self.schema_hash = schema.get_hash()

        self._full_graphql_schema: GraphQLSchema | None = None
        self._graphql_types: dict[str, GraphQLTypes] = {}

        self._load_attribute_types()
        self._load_event_types()
        if config.SETTINGS.experimental_features.graphql_enums:
            self._load_all_enum_types(node_schemas=self.schema.get_all().values())
        self._load_node_interface()

    def get_graphql_types(self) -> dict[str, GraphQLTypes]:
        return self._graphql_types

    def get_graphql_schema(
        self,
        include_query: bool = True,
        include_mutation: bool = True,
        include_subscription: bool = True,
        include_types: bool = True,
    ) -> GraphQLSchema:
        if all((include_query, include_mutation, include_subscription, include_types)):
            if not self._full_graphql_schema:
                self._full_graphql_schema = self.generate()
            return self._full_graphql_schema
        return self.generate(
            include_query=include_query,
            include_mutation=include_mutation,
            include_subscription=include_subscription,
            include_types=include_types,
        )

    def generate(
        self,
        include_query: bool = True,
        include_mutation: bool = True,
        include_subscription: bool = True,
        include_types: bool = True,
    ) -> GraphQLSchema:
        with SCHEMA_GENERATE_GRAPHQL_METRICS.labels(self.schema.name).time():
            query = self.get_gql_query() if include_query else None

            if include_types:
                if not include_query:
                    self.generate_object_types()
                types_dict = self.get_all()
                types = list(types_dict.values())
            else:
                types = []

            mutation = self.get_gql_mutation() if include_mutation else None
            subscription = None
            if include_subscription:
                partial_graphene_schema = graphene.Schema(
                    query=query,
                    mutation=mutation,
                    types=types,
                    auto_camelcase=False,
                    directives=DIRECTIVES,
                )
                subscription = self.get_gql_subscription(partial_graphql_schema=partial_graphene_schema.graphql_schema)

            graphene_schema = graphene.Schema(
                query=query,
                mutation=mutation,
                subscription=subscription,
                types=types,
                auto_camelcase=False,
                directives=DIRECTIVES,
            )

            return graphene_schema.graphql_schema

    def get_gql_query(self) -> type[InfrahubBaseQuery]:
        QueryMixin = self.generate_query_mixin()

        class Query(InfrahubBaseQuery, QueryMixin):  # type: ignore
            pass

        return Query

    def get_gql_mutation(self) -> type[InfrahubBaseMutation]:
        MutationMixin = self.generate_mutation_mixin()

        class Mutation(InfrahubBaseMutation, MutationMixin):  # type: ignore
            pass

        return Mutation

    def get_gql_subscription(self, partial_graphql_schema: graphene.Schema) -> type[InfrahubBaseSubscription]:
        class Subscription(InfrahubBaseSubscription):
            graphql_schema = partial_graphql_schema

        return Subscription

    def get_type(self, name: str) -> type[InfrahubObject]:
        if name in self._graphql_types and issubclass(
            self._graphql_types[name], BaseAttributeType | graphene.Interface | graphene.ObjectType
        ):
            return self._graphql_types[name]
        raise ValueError(f"Unable to find {name!r}")

    def get_mutation(self, name: str) -> type[InfrahubMutation]:
        if name in self._graphql_types and issubclass(self._graphql_types[name], InfrahubMutation):
            return self._graphql_types[name]
        raise ValueError(f"Unable to find {name!r}")

    def get_all(self) -> dict[str, GraphQLTypes]:
        return self._graphql_types

    def set_type(self, name: str, graphql_type: GraphQLTypes) -> None:
        self._graphql_types[name] = graphql_type

    def _load_attribute_types(self) -> None:
        for data_type in ATTRIBUTE_TYPES.values():
            self.set_type(name=data_type.get_graphql_type_name(), graphql_type=data_type.get_graphql_type())

    def _load_event_types(self) -> None:
        for event in EVENT_TYPES.values():
            self.set_type(name=event._meta.name, graphql_type=event)

    def _load_node_interface(self) -> None:
        """Load the base CoreNode interface. Edged/paginated objects are created later in generate_object_types."""
        node_interface_schema = GenericSchema(
            name="Node", namespace="Core", description="Interface for all nodes in Infrahub"
        )
        self.generate_interface_object(schema=node_interface_schema, populate_cache=True)

    def _complete_node_interface(self, node_metadata: type[InfrahubObject]) -> None:
        """Complete the CoreNode interface by creating its edged and paginated objects."""
        node_interface_schema = GenericSchema(
            name="Node", namespace="Core", description="Interface for all nodes in Infrahub"
        )
        # Re-call generate_interface_object to get the InterfaceReference (will use cached version)
        interface = self.generate_interface_object(schema=node_interface_schema, populate_cache=False)
        edged_interface = self.generate_graphql_edged_object(
            schema=node_interface_schema, node=interface, node_metadata=node_metadata, populate_cache=True
        )
        self.generate_graphql_paginated_object(schema=node_interface_schema, edge=edged_interface, populate_cache=True)

    def _patch_static_types(self, node_metadata: type[InfrahubObject]) -> None:
        """Patch statically defined GraphQL types to use dynamically generated types.

        Some GraphQL types like InfrahubBranchEdge are defined statically but need to
        reference dynamically generated types (like node_metadata with GenericAccount).
        This method patches those static types after the dynamic types are created.

        The method checks if the patch has already been applied to avoid redundant updates.
        """
        current_field = InfrahubBranchEdge._meta.fields.get("node_metadata")
        if current_field is None or current_field.type != node_metadata:
            InfrahubBranchEdge._meta.fields["node_metadata"] = graphene.Field(node_metadata, required=True)

    def _load_all_enum_types(self, node_schemas: Iterable[MainSchemaTypes]) -> None:
        for node_schema in node_schemas:
            self._load_enum_type(node_schema=node_schema)

    def _load_enum_type(self, node_schema: MainSchemaTypes) -> None:
        for attr_schema in node_schema.attributes:
            if not attr_schema.enum:
                continue
            base_enum_name = get_enum_attribute_type_name(node_schema, attr_schema)
            enum_value_name = f"{base_enum_name}Value"
            graphene_enum = generate_graphql_enum(name=enum_value_name, options=attr_schema.enum)
            data_type_class_name = f"{base_enum_name}EnumType"

            default_value = None
            if attr_schema.default_value:
                for g_enum in graphene_enum:
                    if g_enum.value == attr_schema.default_value:
                        default_value = g_enum.name
                        break

            graphene_field = graphene.Field(graphene_enum, default_value=default_value)
            create_class = type(f"{base_enum_name}AttributeCreate", (BaseAttributeCreate,), {"value": graphene_field})
            update_class = type(f"{base_enum_name}AttributeUpdate", (BaseAttributeUpdate,), {"value": graphene_field})
            data_type_class: type[InfrahubDataType] = type(
                data_type_class_name,
                (InfrahubDataType,),
                {
                    "label": data_type_class_name,
                    "graphql": graphene.String,
                    "graphql_query": TextAttributeType,
                    "graphql_create": create_class,
                    "graphql_update": update_class,
                    "graphql_filter": graphene_enum,
                    "infrahub": String,
                },
            )
            self.set_type(
                name=data_type_class.get_graphql_type_name(),
                graphql_type=data_type_class.get_graphql_type(),
            )
            ATTRIBUTE_TYPES[base_enum_name] = data_type_class

    def _get_related_input_type(
        self, relationship: RelationshipSchema
    ) -> type[RelatedNodeInput | RelatedIPPrefixNodeInput | RelatedIPAddressNodeInput]:
        peer_schema = self.schema.get(name=relationship.peer, duplicate=False)
        if peer_schema.is_ip_prefix:
            return RelatedIPPrefixNodeInput

        if peer_schema.is_ip_address:
            return RelatedIPAddressNodeInput

        return RelatedNodeInput

    def generate_object_types(self) -> None:
        """Generate all GraphQL objects for the schema and store them in the internal registry."""

        full_schema = self.schema.get_all(duplicate=False)

        # Pass 1: Generate all GraphQL Interface objects first (without edged/paginated)
        # This ensures GENERICACCOUNT exists before we create node_metadata
        for node_schema in full_schema.values():
            if isinstance(node_schema, GenericSchema):
                self.generate_interface_object(schema=node_schema, populate_cache=True)

        # Define LineageSource and LineageOwner
        data_source = self.get_type(name=InfrahubKind.LINEAGESOURCE)
        data_owner = self.get_type(name=InfrahubKind.LINEAGEOWNER)
        self.define_relationship_property(data_source=data_source, data_owner=data_owner)

        # Now that GENERICACCOUNT exists, create node_metadata and relationship_metadata
        account_type = self.get_type(name=InfrahubKind.GENERICACCOUNT)
        self.define_node_metadata(account_type=account_type)
        self.define_relationship_metadata(account_type=account_type)
        node_metadata = self.get_type(name=NODE_METADATA_TYPE)
        relationship_metadata = self.get_type(name=RELATIONSHIP_METADATA_TYPE)

        # Complete the CoreNode interface (edged/paginated) now that node_metadata exists
        self._complete_node_interface(node_metadata=node_metadata)

        # Patch statically defined types to use dynamically generated types
        self._patch_static_types(node_metadata=node_metadata)

        relationship_property = self.get_type(name="RelationshipProperty")
        for data_type in ATTRIBUTE_TYPES.values():
            gql_type = self.get_type(name=data_type.get_graphql_type_name())
            gql_type._meta.fields["source"] = graphene.Field(data_source)
            gql_type._meta.fields["owner"] = graphene.Field(data_owner)
            gql_type._meta.fields["updated_by"] = graphene.Field(
                account_type, required=False, resolver=account_metadata_resolver
            )

        # Pass 2: Generate edged/paginated objects for all GenericSchema interfaces
        for node_schema in full_schema.values():
            if not isinstance(node_schema, GenericSchema):
                continue
            # Re-call generate_interface_object to get the InterfaceReference (will use cached version)
            interface = self.generate_interface_object(schema=node_schema, populate_cache=False)
            edged_interface = self.generate_graphql_edged_object(
                schema=node_schema, node=interface, node_metadata=node_metadata, populate_cache=True
            )
            self.generate_graphql_paginated_object(schema=node_schema, edge=edged_interface, populate_cache=True)

        # Generate all Nested, Edged and NestedEdged Interfaces and store them in the registry
        for node_name, node_schema in full_schema.items():
            if not isinstance(node_schema, GenericSchema):
                continue
            node_interface = self.get_type(name=node_name)

            nested_edged_interface = self.generate_nested_interface_object(
                schema=node_schema,
                base_interface=node_interface,
                node_metadata=node_metadata,
                relation_property=relationship_property,
                relationship_metadata=relationship_metadata,
            )

            nested_interface = self.generate_paginated_interface_object(
                schema=node_schema,
                base_interface=nested_edged_interface,
            )

            self.set_type(name=nested_interface._meta.name, graphql_type=nested_interface)
            self.set_type(name=nested_edged_interface._meta.name, graphql_type=nested_edged_interface)

        # Generate all GraphQL ObjectType, Nested, Paginated & NestedPaginated and store them in the registry
        for node_schema in full_schema.values():
            if isinstance(node_schema, NodeSchema | ProfileSchema | TemplateSchema):
                node_object_type = self.generate_graphql_object(schema=node_schema, populate_cache=True)
                node_type_edged = self.generate_graphql_edged_object(
                    schema=node_schema, node=node_object_type, node_metadata=node_metadata, populate_cache=True
                )
                nested_node_type_edged = self.generate_graphql_edged_object(
                    schema=node_schema,
                    node=node_object_type,
                    node_metadata=node_metadata,
                    relationship_metadata=relationship_metadata,
                    relation_property=relationship_property,
                    populate_cache=True,
                )

                self.generate_graphql_paginated_object(schema=node_schema, edge=node_type_edged, populate_cache=True)
                self.generate_graphql_paginated_object(
                    schema=node_schema, edge=nested_node_type_edged, nested=True, populate_cache=True
                )

        # Extend all types and related types with Relationships
        for node_name, node_schema in full_schema.items():
            node_type = self.get_type(name=node_name)

            for rel in node_schema.relationships:
                # Exclude hierarchical relationships, we will add them later
                if (
                    (isinstance(node_schema, NodeSchema) and node_schema.hierarchy)
                    or (isinstance(node_schema, GenericSchema) and node_schema.hierarchical)
                ) and rel.name in ("parent", "children", "ancestors", "descendants"):
                    continue

                peer_schema = self.schema.get(name=rel.peer, duplicate=False)
                if peer_schema.namespace == "Internal":
                    continue
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False)

                if rel.cardinality == RelationshipCardinality.ONE:
                    peer_type = self.get_type(name=f"NestedEdged{peer_schema.kind}")
                    node_type._meta.fields[rel.name] = graphene.Field(
                        peer_type, resolver=single_relationship_resolver, required=True
                    )

                elif rel.cardinality == RelationshipCardinality.MANY:
                    peer_type = self.get_type(name=f"NestedPaginated{peer_schema.kind}")

                    if (isinstance(node_schema, NodeSchema) and node_schema.hierarchy) or (
                        isinstance(node_schema, GenericSchema) and node_schema.hierarchical
                    ):
                        peer_filters["include_descendants"] = graphene.Boolean()

                    node_type._meta.fields[rel.name] = graphene.Field(
                        peer_type, required=True, resolver=many_relationship_resolver, **peer_filters
                    )

            if (isinstance(node_schema, NodeSchema) and node_schema.hierarchy) or (
                isinstance(node_schema, GenericSchema) and node_schema.hierarchical
            ):
                if isinstance(node_schema, NodeSchema):
                    schema = self.schema.get(name=node_schema.hierarchy, duplicate=False)  # type: ignore[arg-type]
                    hierarchy_name = node_schema.hierarchy
                else:
                    schema = node_schema
                    hierarchy_name = node_schema.kind

                peer_filters = self.generate_filters(schema=schema, top_level=False)
                peer_type = self.get_type(name=f"NestedPaginated{hierarchy_name}")
                peer_type_edge = self.get_type(name=f"NestedEdged{hierarchy_name}")

                node_type._meta.fields["parent"] = graphene.Field(
                    peer_type_edge, required=True, resolver=single_relationship_resolver
                )
                node_type._meta.fields["children"] = graphene.Field(
                    peer_type, required=True, resolver=many_relationship_resolver, **peer_filters
                )
                node_type._meta.fields["ancestors"] = graphene.Field(
                    peer_type, required=True, resolver=ancestors_resolver, **peer_filters
                )
                node_type._meta.fields["descendants"] = graphene.Field(
                    peer_type, required=True, resolver=descendants_resolver, **peer_filters
                )

    def generate_query_mixin(self) -> type[object]:
        class_attrs = {}

        full_schema = self.schema.get_all(duplicate=False)

        # Generate all Graphql objectType and store internally
        self.generate_object_types()

        for node_name, node_schema in full_schema.items():
            if node_schema.namespace == "Internal":
                continue

            node_type = self.get_type(name=f"Paginated{node_name}")
            node_filters = self.generate_filters(schema=node_schema, top_level=True)

            class_attrs[node_schema.kind] = graphene.Field(
                node_type,
                resolver=default_paginated_list_resolver
                if node_name not in [InfrahubKind.IPADDRESS, InfrahubKind.IPPREFIX]
                else ipam_paginated_list_resolver,
                required=True,
                **node_filters,
            )
            if node_name == InfrahubKind.GENERICACCOUNT:
                node_type = self.get_type(name=InfrahubKind.GENERICACCOUNT)
                class_attrs["AccountProfile"] = graphene.Field(node_type, resolver=account_resolver)

        return type("QueryMixin", (object,), class_attrs)

    def generate_mutation_mixin(self) -> type[object]:
        class_attrs: dict[str, Any] = {}

        full_schema = self.schema.get_all(duplicate=False)

        for node_schema in full_schema.values():
            if node_schema.namespace == "Internal":
                continue

            mutation_map: dict[str, type[InfrahubMutation]] = {
                InfrahubKind.ARTIFACTDEFINITION: InfrahubArtifactDefinitionMutation,
                InfrahubKind.REPOSITORY: InfrahubRepositoryMutation,
                InfrahubKind.READONLYREPOSITORY: InfrahubRepositoryMutation,
                InfrahubKind.PROPOSEDCHANGE: InfrahubProposedChangeMutation,
                InfrahubKind.GRAPHQLQUERY: InfrahubGraphQLQueryMutation,
                InfrahubKind.NAMESPACE: InfrahubIPNamespaceMutation,
                InfrahubKind.NUMBERPOOL: InfrahubNumberPoolMutation,
                InfrahubKind.MENUITEM: InfrahubCoreMenuMutation,
                InfrahubKind.STANDARDWEBHOOK: InfrahubWebhookMutation,
                InfrahubKind.CUSTOMWEBHOOK: InfrahubWebhookMutation,
                InfrahubKind.NODETRIGGERRULE: InfrahubTriggerRuleMutation,
                InfrahubKind.NODETRIGGERATTRIBUTEMATCH: InfrahubTriggerRuleMatchMutation,
                InfrahubKind.NODETRIGGERRELATIONSHIPMATCH: InfrahubTriggerRuleMatchMutation,
            }

            if isinstance(node_schema, NodeSchema) and node_schema.is_ip_prefix:
                base_class = InfrahubIPPrefixMutation
            elif isinstance(node_schema, NodeSchema) and node_schema.is_ip_address:
                base_class = InfrahubIPAddressMutation
            elif isinstance(node_schema, ProfileSchema):
                base_class = InfrahubProfileMutation
            else:
                base_class = mutation_map.get(node_schema.kind, InfrahubMutation)

            if isinstance(node_schema, NodeSchema | ProfileSchema | TemplateSchema):
                mutations = self.generate_graphql_mutations(schema=node_schema, base_class=base_class)

                class_attrs[f"{node_schema.kind}Create"] = mutations.create.Field()
                class_attrs[f"{node_schema.kind}Update"] = mutations.update.Field()
                class_attrs[f"{node_schema.kind}Upsert"] = mutations.upsert.Field()
                class_attrs[f"{node_schema.kind}Delete"] = mutations.delete.Field()

            elif (
                isinstance(node_schema, GenericSchema)
                and (len(node_schema.attributes) + len(node_schema.relationships)) > 0
            ):
                graphql_mutation_update_input = self.generate_graphql_mutation_update_input(node_schema)
                update = self.generate_graphql_mutation_update(
                    schema=node_schema, base_class=base_class, input_type=graphql_mutation_update_input
                )
                self.set_type(name=update._meta.name, graphql_type=update)
                class_attrs[f"{node_schema.kind}Update"] = update.Field()

        return type("MutationMixin", (object,), class_attrs)

    def generate_graphql_object(self, schema: MainSchemaTypes, populate_cache: bool = False) -> InfrahubObjectReference:
        """Generate a GraphQL object Type from a Infrahub NodeSchema."""

        interfaces: set[type[InfrahubObject]] = set()
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{schema.kind}{schema.get_hash()}".encode())

        if isinstance(schema, NodeSchema | ProfileSchema | TemplateSchema) and schema.inherit_from:
            for generic_name in sorted(schema.inherit_from):
                generic = self.get_type(name=generic_name)
                interfaces.add(generic)

        if isinstance(schema, NodeSchema):
            if not schema.inherit_from or InfrahubKind.GENERICGROUP not in schema.inherit_from:
                node_interface = self.get_type(name=InfrahubKind.NODE)
                interfaces.add(node_interface)

        meta_attrs = {
            "schema": schema,
            "name": schema.kind,
            "description": schema.description,
            "interfaces": interfaces,
        }

        main_attrs = {
            "id": graphene.Field(graphene.String, required=True, description="Unique identifier"),
            "hfid": graphene.Field(
                graphene.List(of_type=graphene.NonNull(graphene.String)),
                required=False,
                description="Human friendly identifier",
            ),
            "_updated_at": graphene.DateTime(
                required=False,
                deprecation_reason="Query the node_metadata field instead. Will be removed in Infrahub 1.9",
            ),
            "display_label": graphene.String(required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        for attr in schema.local_attributes:
            attr_kind = get_attr_kind(schema, attr)
            attr_type = self.get_type(name=get_attribute_type(kind=attr_kind).get_graphql_type_name())
            req = "" if attr.optional else " (required)"
            main_attrs[attr.name] = graphene.Field(attr_type, description=f"{attr.description}{req}")

        object_hash = md5hash.hexdigest()

        graphql_object = type(schema.kind, (InfrahubObject,), main_attrs)
        registry.set_object_type(reference=graphql_object, reference_hash=object_hash, schema_hash=self.schema_hash)

        if populate_cache:
            self.set_type(name=schema.kind, graphql_type=graphql_object)

        return InfrahubObjectReference(reference=graphql_object, reference_hash=object_hash)

    def generate_interface_object(self, schema: GenericSchema, populate_cache: bool = False) -> InterfaceReference:
        meta_attrs = {
            "name": schema.kind,
            "description": schema.description,
        }

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"interface-{schema.kind}{schema.get_hash()}".encode())
        interface_hash = md5hash.hexdigest()

        main_attrs = {
            "id": graphene.Field(graphene.String, required=False, description="Unique identifier"),
            "hfid": graphene.Field(
                graphene.List(of_type=graphene.NonNull(graphene.String)),
                required=False,
                description="Human friendly identifier",
            ),
            "display_label": graphene.String(required=False),
        }

        for attr in schema.attributes:
            attr_kind = get_attr_kind(node_schema=schema, attr_schema=attr)
            attr_type = self.get_type(name=get_attribute_type(kind=attr_kind).get_graphql_type_name())
            main_attrs[attr.name] = graphene.Field(attr_type, description=attr.description)

        interface_object = registry.get_interface_type(reference_hash=interface_hash, schema_hash=self.schema_hash)
        if not interface_object:
            main_attrs["Meta"] = type("Meta", (object,), meta_attrs)
            interface_object = type(schema.kind, (InfrahubInterface,), main_attrs)

        interface_reference = InterfaceReference(reference=interface_object, reference_hash=interface_hash)
        registry.set_interface_type(reference=interface_reference, schema_hash=self.schema_hash)

        if populate_cache:
            self.set_type(name=schema.kind, graphql_type=interface_object)

        return interface_reference

    def define_relationship_property(self, data_source: type[InfrahubObject], data_owner: type[InfrahubObject]) -> None:
        type_name = "RelationshipProperty"

        meta_attrs = {
            "name": type_name,
            "description": "Defines properties for relationships",
        }

        main_attrs = {
            "is_protected": graphene.Boolean(required=False),
            "updated_at": graphene.DateTime(required=False),
            "source": graphene.Field(data_source),
            "owner": graphene.Field(data_owner),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        relationship_property = type(type_name, (graphene.ObjectType,), main_attrs)

        self.set_type(name=type_name, graphql_type=relationship_property)

    def define_node_metadata(self, account_type: type[InfrahubObject]) -> None:
        meta_attrs = {
            "name": NODE_METADATA_TYPE,
            "description": "Defines node metadata information",
        }

        main_attrs = {
            "created_at": graphene.DateTime(required=False),
            "created_by": graphene.Field(account_type, required=False, resolver=account_metadata_resolver),
            "updated_at": graphene.DateTime(required=False),
            "updated_by": graphene.Field(account_type, required=False, resolver=account_metadata_resolver),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        node_metadata = type(NODE_METADATA_TYPE, (graphene.ObjectType,), main_attrs)

        self.set_type(name=NODE_METADATA_TYPE, graphql_type=node_metadata)

    def define_relationship_metadata(self, account_type: type[InfrahubObject]) -> None:
        meta_attrs = {
            "name": RELATIONSHIP_METADATA_TYPE,
            "description": "Defines relationship metadata information",
        }

        main_attrs = {
            "created_at": graphene.DateTime(required=False),
            "created_by": graphene.Field(account_type, required=False, resolver=account_metadata_resolver),
            "updated_at": graphene.DateTime(required=False),
            "updated_by": graphene.Field(account_type, required=False, resolver=account_metadata_resolver),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        relationship_metadata = type(RELATIONSHIP_METADATA_TYPE, (graphene.ObjectType,), main_attrs)

        self.set_type(name=RELATIONSHIP_METADATA_TYPE, graphql_type=relationship_metadata)

    def generate_graphql_mutations(
        self,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        base_class: type[InfrahubMutation],
    ) -> GraphqlMutations:
        graphql_mutation_create_input = self.generate_graphql_mutation_create_input(schema)
        graphql_mutation_update_input = self.generate_graphql_mutation_update_input(schema)
        graphql_mutation_upsert_input = self.generate_graphql_mutation_upsert_input(schema)

        create = self.generate_graphql_mutation_create(
            schema=schema, base_class=base_class, input_type=graphql_mutation_create_input
        )
        upsert = self.generate_graphql_mutation_create(
            schema=schema,
            base_class=base_class,
            input_type=graphql_mutation_upsert_input,
            mutation_type="Upsert",
        )
        update = self.generate_graphql_mutation_update(
            schema=schema, base_class=base_class, input_type=graphql_mutation_update_input
        )
        delete = self.generate_graphql_mutation_delete(schema=schema, base_class=base_class)

        self.set_type(name=create._meta.name, graphql_type=create)
        self.set_type(name=update._meta.name, graphql_type=update)
        self.set_type(name=upsert._meta.name, graphql_type=upsert)
        self.set_type(name=delete._meta.name, graphql_type=delete)

        return GraphqlMutations(create=create, update=update, upsert=upsert, delete=delete)

    def generate_graphql_mutation_create_input(
        self, schema: NodeSchema | ProfileSchema | TemplateSchema
    ) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusCreateInput(InputObjectType):
                id = String(required=False)
                label = InputField(StringAttributeCreate, required=True)
                slug = InputField(StringAttributeCreate, required=True)
                description = InputField(StringAttributeCreate, required=False)
        """
        attrs: dict[str, graphene.String | graphene.InputField] = {"id": graphene.String(required=False)}

        for attr in schema.attributes:
            if attr.read_only:
                continue

            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_create()

            attrs[attr.name] = graphene.InputField(attr_type, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            if rel.cardinality == RelationshipCardinality.ONE:
                attrs[rel.name] = graphene.InputField(input_type, description=rel.description)

            elif rel.cardinality == RelationshipCardinality.MANY:
                attrs[rel.name] = graphene.InputField(graphene.List(input_type), description=rel.description)

        input_name = f"{schema.kind}CreateInput"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{input_name}{schema.get_hash()}".encode())
        input_hash = md5hash.hexdigest()
        mutation_input_type = registry.get_input_type(reference_hash=input_hash, schema_hash=self.schema_hash)
        if not mutation_input_type:
            mutation_input_type = type(input_name, (graphene.InputObjectType,), attrs)
            registry.set_input_type(
                reference=mutation_input_type, reference_hash=input_hash, schema_hash=self.schema_hash
            )

        return mutation_input_type

    def generate_graphql_mutation_update_input(self, schema: MainSchemaTypes) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusUpdateInput(InputObjectType):
                id = String(required=False)
                hfid = InputField(List(of_type=String), required=False)
                label = InputField(StringAttributeUpdate, required=False)
                slug = InputField(StringAttributeUpdate, required=False)
                description = InputField(StringAttributeUpdate, required=False)
        """
        attrs: dict[str, graphene.String | graphene.InputField | graphene.List] = {
            "id": graphene.String(required=False),
            "hfid": graphene.List(of_type=graphene.String, required=False),
        }

        for attr in schema.attributes:
            if attr.read_only:
                continue
            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_update()
            attrs[attr.name] = graphene.InputField(attr_type, required=False, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            if rel.cardinality == RelationshipCardinality.ONE:
                attrs[rel.name] = graphene.InputField(input_type, required=False, description=rel.description)

            elif rel.cardinality == RelationshipCardinality.MANY:
                attrs[rel.name] = graphene.InputField(
                    graphene.List(input_type), required=False, description=rel.description
                )

        input_name = f"{schema.kind}UpdateInput"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{input_name}{schema.get_hash()}".encode())
        input_hash = md5hash.hexdigest()
        mutation_input_type = registry.get_input_type(reference_hash=input_hash, schema_hash=self.schema_hash)
        if not mutation_input_type:
            mutation_input_type = type(input_name, (graphene.InputObjectType,), attrs)
            registry.set_input_type(
                reference=mutation_input_type, reference_hash=input_hash, schema_hash=self.schema_hash
            )

        return mutation_input_type

    def generate_graphql_mutation_upsert_input(
        self, schema: NodeSchema | ProfileSchema | TemplateSchema
    ) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusUpsertInput(InputObjectType):
                id = String(required=False)
                hfid = InputField(List(of_type=String), required=False)
                label = InputField(StringAttributeUpdate, required=True)
                slug = InputField(StringAttributeUpdate, required=True)
                description = InputField(StringAttributeUpdate, required=False)
        """
        attrs: dict[str, graphene.String | graphene.InputField | graphene.List] = {
            "id": graphene.String(required=False),
            "hfid": graphene.List(of_type=graphene.String, required=False),
        }

        for attr in schema.attributes:
            if attr.read_only:
                continue

            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_update()

            attrs[attr.name] = graphene.InputField(attr_type, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            if rel.cardinality == RelationshipCardinality.ONE:
                attrs[rel.name] = graphene.InputField(input_type, description=rel.description)

            elif rel.cardinality == RelationshipCardinality.MANY:
                attrs[rel.name] = graphene.InputField(graphene.List(input_type), description=rel.description)

        input_name = f"{schema.kind}UpsertInput"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{input_name}{schema.get_hash()}".encode())
        input_hash = md5hash.hexdigest()
        mutation_input_type = registry.get_input_type(reference_hash=input_hash, schema_hash=self.schema_hash)
        if not mutation_input_type:
            mutation_input_type = type(input_name, (graphene.InputObjectType,), attrs)
            registry.set_input_type(
                reference=mutation_input_type, reference_hash=input_hash, schema_hash=self.schema_hash
            )

        return mutation_input_type

    def generate_graphql_mutation_create(
        self,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        input_type: type[graphene.InputObjectType],
        base_class: type[InfrahubMutation] = InfrahubMutation,
        mutation_type: str = "Create",
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to CREATE an object based on the specified NodeSchema."""
        name = f"{schema.kind}{mutation_type}"

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{name}{schema.get_hash()}".encode())
        mutation_hash = md5hash.hexdigest()

        mutation_object = registry.get_mutation_type(reference_hash=mutation_hash, schema_hash=self.schema_hash)
        if not mutation_object:
            object_type = self.generate_graphql_object(schema=schema)

            main_attrs: dict[str, Any] = {"ok": graphene.Boolean(), "object": graphene.Field(object_type.reference)}

            meta_attrs: dict[str, Any] = {"schema": schema, "name": name, "description": schema.description}
            main_attrs["Meta"] = type("Meta", (object,), meta_attrs)

            args_attrs: dict[str, Any] = {"data": input_type(required=True), "context": ContextInput(required=False)}
            if schema.is_file_object:
                args_attrs["file"] = graphene.Argument(Upload, required=True)

            main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)
            mutation_object = type(name, (base_class,), main_attrs)
            registry.set_mutation_type(
                reference=mutation_object, reference_hash=mutation_hash, schema_hash=self.schema_hash
            )

        return mutation_object

    def generate_graphql_mutation_update(
        self,
        schema: MainSchemaTypes,
        input_type: type[graphene.InputObjectType],
        base_class: type[InfrahubMutation] = InfrahubMutation,
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to UPDATE an object based on the specified NodeSchema."""
        name = f"{schema.kind}Update"

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{name}{schema.get_hash()}".encode())
        mutation_hash = md5hash.hexdigest()

        mutation_object = registry.get_mutation_type(reference_hash=mutation_hash, schema_hash=self.schema_hash)
        if not mutation_object:
            object_type = self.generate_graphql_object(schema=schema)

            main_attrs: dict[str, Any] = {"ok": graphene.Boolean(), "object": graphene.Field(object_type.reference)}

            meta_attrs: dict[str, Any] = {"schema": schema, "name": name, "description": schema.description}
            main_attrs["Meta"] = type("Meta", (object,), meta_attrs)

            args_attrs: dict[str, Any] = {"data": input_type(required=True), "context": ContextInput(required=False)}
            if schema.is_file_object:
                args_attrs["file"] = graphene.Argument(Upload, required=False)

            main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)

            mutation_object = type(name, (base_class,), main_attrs)
            registry.set_mutation_type(
                reference=mutation_object, reference_hash=mutation_hash, schema_hash=self.schema_hash
            )

        return mutation_object

    def generate_graphql_mutation_delete(
        self, schema: NodeSchema | ProfileSchema | TemplateSchema, base_class: type[InfrahubMutation] = InfrahubMutation
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to DELETE an object based on the specified NodeSchema."""
        name = f"{schema.kind}Delete"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{name}{schema.get_hash()}".encode())
        mutation_hash = md5hash.hexdigest()
        mutation_object = registry.get_mutation_type(reference_hash=mutation_hash, schema_hash=self.schema_hash)
        if not mutation_object:
            main_attrs: dict[str, Any] = {"ok": graphene.Boolean()}
            meta_attrs = {"schema": schema, "name": name, "description": schema.description}
            main_attrs["Meta"] = type("Meta", (object,), meta_attrs)
            args_attrs: dict[str, Any] = {"data": DeleteInput(required=True), "context": ContextInput(required=False)}
            main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)
            mutation_object = type(name, (base_class,), main_attrs)
            registry.set_mutation_type(
                reference=mutation_object, reference_hash=mutation_hash, schema_hash=self.schema_hash
            )

        return mutation_object

    def generate_filters(
        self, schema: MainSchemaTypes, top_level: bool = False, include_properties: bool = True
    ) -> dict[str, graphene.Scalar | graphene.List]:
        """Generate the GraphQL filters for a given Schema object.

        The generated filter will be different if we are at the top_level (query)
        or if we are generating the filter for a relationship inside a node.

        At the top, level it will be possible to query with a list of ID
        Inside a node, it's only possible to query with a single ID

        Args:
            schema (Union[NodeSchema, GenericSchema]): Schema to generate the filters
            top_level (bool, optional): Flag to indicate if are at the top level or not. Defaults to False.

        Returns:
            dict: A Dictionary containing all the filters with their name as the key and their Type as value
        """

        filters: dict[str, Any] = {"offset": graphene.Int(), "limit": graphene.Int(), "order": OrderInput()}
        default_filters: list[str] = list(filters.keys())

        filters["ids"] = graphene.List(graphene.ID)
        if not top_level:
            filters["isnull"] = graphene.Boolean()

        if schema.display_label:
            display_label_schema = schema.get_attribute("display_label")
            filters.update(
                get_attribute_type(kind=display_label_schema.kind).get_graphql_filters(
                    name="display_label", include_properties=False, include_isnull=True
                )
            )

        if schema.human_friendly_id and top_level:
            # NOTE: this can loosen to allow filtering at a non-top level once IFC-2110 is implemented
            filters["hfid"] = graphene.List(graphene.String)

        for attr in schema.attributes:
            attr_kind = get_attr_kind(node_schema=schema, attr_schema=attr)
            filters.update(
                get_attribute_type(kind=attr_kind).get_graphql_filters(
                    name=attr.name, include_properties=include_properties, include_isnull=top_level
                )
            )

        if top_level:
            filters.update(get_attribute_type().get_graphql_filters(name="any"))
            filters["partial_match"] = graphene.Boolean()

            # Add metadata filters for filtering by created_by, updated_by, created_at, updated_at
            filters.update(self._generate_metadata_filters())

            if schema.kind in [InfrahubKind.IPADDRESS, InfrahubKind.IPPREFIX]:
                # This is only available for IPAM generics
                filters["include_available"] = graphene.Boolean()
                filters["kinds"] = graphene.List(graphene.NonNull(graphene.String))

        if not top_level:
            return filters

        for rel in schema.relationships:
            peer_schema = self.schema.get(name=rel.peer, duplicate=False)

            if peer_schema.namespace == "Internal":
                continue

            if rel.kind == RelationshipKind.GROUP:
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False, include_properties=False)
            else:
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False)

            for key, value in peer_filters.items():
                if key in default_filters:
                    continue
                filters[f"{rel.name}__{key}"] = value

        return filters

    def _generate_metadata_filters(self) -> dict[str, Any]:
        """Generate GraphQL filters for object-level metadata fields.

        These filters allow querying nodes based on their metadata:
        - created_by: Filter by the account that created the node
        - updated_by: Filter by the account that last updated the node
        - created_at: Filter by creation timestamp
        - updated_at: Filter by last update timestamp

        Returns:
            dict: Filter definitions with names as keys and graphene types as values
        """
        return {
            # Account-based filters (created_by)
            "node_metadata__created_by__id": graphene.ID(description="Filter by exact creator account UUID"),
            "node_metadata__created_by__ids": graphene.List(
                graphene.ID, description="Filter by list of creator account UUIDs"
            ),
            # Account-based filters (updated_by)
            "node_metadata__updated_by__id": graphene.ID(description="Filter by exact updater account UUID"),
            "node_metadata__updated_by__ids": graphene.List(
                graphene.ID, description="Filter by list of updater account UUIDs"
            ),
            # DateTime-based filters (created_at)
            "node_metadata__created_at": graphene.DateTime(description="Filter by exact creation timestamp"),
            "node_metadata__created_at__before": graphene.DateTime(
                description="Filter for objects created before this timestamp"
            ),
            "node_metadata__created_at__after": graphene.DateTime(
                description="Filter for objects created after this timestamp"
            ),
            # DateTime-based filters (updated_at)
            "node_metadata__updated_at": graphene.DateTime(description="Filter by exact update timestamp"),
            "node_metadata__updated_at__before": graphene.DateTime(
                description="Filter for objects updated before this timestamp"
            ),
            "node_metadata__updated_at__after": graphene.DateTime(
                description="Filter for objects updated after this timestamp"
            ),
        }

    def generate_graphql_edged_object(
        self,
        schema: MainSchemaTypes,
        node: InterfaceReference | InfrahubObjectReference,
        node_metadata: type[InfrahubObject],
        relationship_metadata: type[InfrahubObject] | None = None,
        relation_property: type[InfrahubObject] | None = None,
        populate_cache: bool = False,
    ) -> InfrahubEdgedReference:
        """Generate a edged GraphQL object Type from a Infrahub NodeSchema for pagination."""

        object_name = f"Edged{schema.kind}"
        if relation_property:
            object_name = f"NestedEdged{schema.kind}"

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{object_name}{schema.get_hash()}{node.reference_hash}".encode())
        edge_hash = md5hash.hexdigest()

        meta_attrs: dict[str, Any] = {
            "schema": schema,
            "name": object_name,
            "description": schema.description,
            "interfaces": set(),
        }

        main_attrs: dict[str, Any] = {
            "node": graphene.Field(node.reference, required=False),
            "node_metadata": graphene.Field(node_metadata, required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        if relation_property:
            main_attrs["properties"] = graphene.Field(relation_property, required=False)
        if relationship_metadata:
            main_attrs["relationship_metadata"] = graphene.Field(relationship_metadata, required=False)

        graphql_edged_object = registry.get_edge_type(reference_hash=edge_hash, schema_hash=self.schema_hash)
        if not graphql_edged_object:
            graphql_edged_object = type(object_name, (InfrahubObjectWithoutMeta,), main_attrs)
            registry.set_edge_type(
                reference=graphql_edged_object, reference_hash=edge_hash, schema_hash=self.schema_hash
            )

        if populate_cache:
            self.set_type(name=object_name, graphql_type=graphql_edged_object)

        return InfrahubEdgedReference(reference=graphql_edged_object, reference_hash=edge_hash)

    def generate_graphql_paginated_object(
        self, schema: MainSchemaTypes, edge: InfrahubEdgedReference, nested: bool = False, populate_cache: bool = False
    ) -> type[InfrahubObject]:
        """Generate a paginated GraphQL object Type from a Infrahub NodeSchema."""

        object_name = f"Paginated{schema.kind}"
        if nested:
            object_name = f"NestedPaginated{schema.kind}"

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{object_name}{schema.get_hash()}{edge.reference_hash}".encode())
        paginated_hash = md5hash.hexdigest()

        meta_attrs: dict[str, Any] = {
            "schema": schema,
            "name": object_name,
            "description": schema.description,
            "default_resolver": default_resolver,
            "interfaces": set(),
        }

        main_attrs: dict[str, Any] = {
            "count": graphene.Int(required=True),
            "edges": graphene.List(of_type=graphene.NonNull(edge.reference), required=True),
            "permissions": graphene.Field(
                PaginatedObjectPermission, required=True, resolver=parent_field_name_resolver
            ),
        }

        graphql_paginated_object = registry.get_paginated_type(
            reference_hash=paginated_hash, schema_hash=self.schema_hash
        )
        if not graphql_paginated_object:
            main_attrs["Meta"] = type("Meta", (object,), meta_attrs)
            graphql_paginated_object = type(object_name, (InfrahubObjectWithoutMeta,), main_attrs)
            registry.set_paginated_type(
                reference=graphql_paginated_object, reference_hash=paginated_hash, schema_hash=self.schema_hash
            )

        if populate_cache:
            self.set_type(name=object_name, graphql_type=graphql_paginated_object)

        return graphql_paginated_object

    def generate_nested_interface_object(
        self,
        schema: GenericSchema,
        relation_property: graphene.ObjectType,
        base_interface: graphene.ObjectType,
        node_metadata: type[InfrahubObject],
        relationship_metadata: type[InfrahubObject] | None = None,
        populate_cache: bool = False,
    ) -> type[InfrahubObject]:
        meta_attrs: dict[str, Any] = {
            "name": f"NestedEdged{schema.kind}",
            "schema": schema,
            "description": schema.description,
        }

        main_attrs: dict[str, Any] = {
            "node": graphene.Field(base_interface, required=False),
            "_updated_at": graphene.DateTime(
                required=False,
                deprecation_reason="Query the node_metadata field instead. Will be removed in Infrahub 1.9",
            ),
            "node_metadata": graphene.Field(node_metadata, required=True),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        if relation_property:
            main_attrs["properties"] = graphene.Field(relation_property, required=False)
        if relationship_metadata:
            main_attrs["relationship_metadata"] = graphene.Field(relationship_metadata, required=False)

        object_name = f"NestedEdged{schema.kind}"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{object_name}{schema.get_hash()}".encode())
        paginated_hash = md5hash.hexdigest()
        nested_interface_object = registry.get_paginated_type(
            reference_hash=paginated_hash, schema_hash=self.schema_hash
        )
        if not nested_interface_object:
            nested_interface_object = type(object_name, (InfrahubObject,), main_attrs)
            registry.set_paginated_type(
                reference=nested_interface_object, reference_hash=paginated_hash, schema_hash=self.schema_hash
            )

        if populate_cache:
            self.set_type(name=object_name, graphql_type=nested_interface_object)

        return nested_interface_object

    def generate_paginated_interface_object(
        self, schema: GenericSchema, base_interface: type[graphene.ObjectType]
    ) -> type[InfrahubObject]:
        object_name = f"NestedPaginated{schema.kind}"
        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(f"{object_name}{schema.get_hash()}".encode())
        paginated_hash = md5hash.hexdigest()

        nested_interface_object = registry.get_paginated_type(
            reference_hash=paginated_hash, schema_hash=self.schema_hash
        )
        if not nested_interface_object:
            meta_attrs: dict[str, Any] = {
                "name": object_name,
                "schema": schema,
                "description": schema.description,
            }

            main_attrs: dict[str, Any] = {
                "count": graphene.Int(required=True),
                "edges": graphene.List(of_type=graphene.NonNull(base_interface)),
                "Meta": type("Meta", (object,), meta_attrs),
            }

            nested_interface_object = type(object_name, (InfrahubObject,), main_attrs)
            registry.set_paginated_type(
                reference=nested_interface_object, reference_hash=paginated_hash, schema_hash=self.schema_hash
            )

        return nested_interface_object


registry._register_manager(manager=GraphQLSchemaManager)



import hashlib
from enum import Enum, EnumMeta
from pathlib import Path
from re import finditer
from typing import Any, TypeVar

KWARGS_TO_DROP = ["session"]
AnyClass = TypeVar("AnyClass", bound=type)


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = Path(__file__).parent.resolve()
    return here.parent / "tests" / "fixtures"


def get_models_dir() -> Path:
    """Get the directory which stores additional models."""
    here = Path(__file__).parent.resolve()
    return here.parent.parent / "models"


def find_first_file_in_directory(directory: Path) -> Path | None:
    return next((f.resolve() for f in directory.iterdir() if f.is_file()), None)


def extract_camelcase_words(camel_case: str) -> list[str]:
    """Extract the namespace and the name for a kind given its camel-case form."""
    matches = finditer(r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)", camel_case)
    return [m.group(0) for m in matches]


def format_label(slug: str) -> str:
    return " ".join([word.title() for word in slug.split("_")])


class MetaEnum(EnumMeta):
    def __contains__(cls, item: Any) -> bool:
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class InfrahubNumberEnum(int, BaseEnum):
    @classmethod
    def available_types(cls) -> list[int]:
        return [cls.__members__[member].value for member in list(cls.__members__)]

    def get_hash(self) -> str:
        return hashlib.md5(str(self.value).encode(), usedforsecurity=False).hexdigest()


class InfrahubStringEnum(str, BaseEnum):
    @classmethod
    def available_types(cls) -> list[str]:
        return [cls.__members__[member].value for member in list(cls.__members__)]

    def get_hash(self) -> str:
        return hashlib.md5(self.value.encode(), usedforsecurity=False).hexdigest()


def get_nested_dict(nested_dict: dict[str, Any], keys: list[str]) -> dict[str, Any]:
    current_level = nested_dict
    for key in keys:
        # Check if the key exists and leads to a dictionary
        if isinstance(current_level, dict) and key in current_level:
            current_level = current_level[key]
        else:
            return {}
    return current_level if isinstance(current_level, dict) else {}


def get_all_subclasses[AnyClass: type](cls: AnyClass) -> list[AnyClass]:
    """Recursively get all subclasses of the given class."""
    subclasses: list[AnyClass] = []
    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(get_all_subclasses(subclass))
    return subclasses


def has_any_key(data: dict[str, Any], keys: list[str]) -> bool:
    """Recursively check if any of the specified keys exist in the dictionary at any level.

    Args:
        data: The dictionary to search through
        keys: List of key names to search for

    Returns:
        True if any of the keys are found at any level of the dictionary, False otherwise
    """
    for key, value in data.items():
        if key in keys:
            return True
        if isinstance(value, dict) and has_any_key(data=value, keys=keys):
            return True
    return False



TEMPLATE_FILE_NAME = "template.j2"

ATTRIBUTE_KIND_MAP = {
    "ID": "String",
    "Text": "String",
    "TextArea": "String",
    "DateTime": "DateTime",
    "Email": "String",
    "Password": "String",
    "HashedPassword": "HashedPassword",
    "URL": "URL",
    "File": "String",
    "MacAddress": "MacAddress",
    "Color": "String",
    "Dropdown": "Dropdown",
    "Number": "Integer",
    "Bandwidth": "Integer",
    "IPHost": "IPHost",
    "IPNetwork": "IPNetwork",
    "Boolean": "Boolean",
    "Checkbox": "Boolean",
    "List": "ListAttribute",
    "JSON": "JSONAttribute",
    "Any": "AnyAttribute",
    "NumberPool": "Integer",
}

# The order of the classes in the list determines the order of the classes in the generated code
CORE_BASE_CLASS_TO_SYNCIFY = ["CoreProfile", "CoreObjectTemplate", "CoreNode"]



from abc import ABC, abstractmethod

from fast_depends import Depends, inject

from infrahub.auth import AccountSession
from infrahub.database import InfrahubDatabase


class BranchMergeMutationChecker(ABC):
    @abstractmethod
    async def verify_branch_merge_mutation_allowed(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,
    ) -> None:
        raise NotImplementedError()


class BranchMergeMutationCheckerCommunity(BranchMergeMutationChecker):
    async def verify_branch_merge_mutation_allowed(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,
    ) -> None:
        pass


def get_branch_merge_mutation_checker() -> BranchMergeMutationChecker:
    return BranchMergeMutationCheckerCommunity()


@inject
async def verify_branch_merge_mutation_allowed(
    db: InfrahubDatabase,
    account_session: AccountSession,
    branch_merge_mutation_checker: BranchMergeMutationChecker = Depends(get_branch_merge_mutation_checker),  # noqa: B008
) -> None:
    await branch_merge_mutation_checker.verify_branch_merge_mutation_allowed(db=db, account_session=account_session)



from dataclasses import dataclass
from enum import StrEnum
from typing import Any

from rich import print as rprint
from rich.console import Console
from rich.table import Table

from infrahub.cli.constants import FAILED_BADGE, SUCCESS_BADGE
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class SchemaFieldType(StrEnum):
    ATTRIBUTE = "attribute"
    RELATIONSHIP = "relationship"


@dataclass
class SchemaFieldDetails:
    schema_kind: str
    schema_uuid: str
    field_type: SchemaFieldType
    field_name: str


class DuplicateSchemaFields(Query):
    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (root:Root)
LIMIT 1
WITH root.default_branch AS default_branch
MATCH (field:SchemaAttribute|SchemaRelationship)
CALL (default_branch, field) {
    MATCH (field)-[is_part_of:IS_PART_OF]->(:Root)
    WHERE is_part_of.branch = default_branch
    ORDER BY is_part_of.from DESC
    RETURN is_part_of
    LIMIT 1
}
WITH default_branch, field, CASE
    WHEN is_part_of.status = "active" AND is_part_of.to IS NULL THEN is_part_of.from
    ELSE NULL
END AS active_from
WHERE active_from IS NOT NULL
WITH default_branch, field, active_from, "SchemaAttribute" IN labels(field) AS is_attribute
CALL (field, default_branch) {
    MATCH (field)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(name_value:AttributeValue)
    WHERE r1.branch = default_branch AND r2.branch = default_branch
    AND r1.status = "active" AND r2.status = "active"
    AND r1.to IS NULL AND r2.to IS NULL
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
    RETURN name_value.value AS field_name
}
CALL (field, default_branch) {
    MATCH (field)-[r1:IS_RELATED]-(rel:Relationship)-[r2:IS_RELATED]-(peer:SchemaNode|SchemaGeneric)
    WHERE rel.name IN ["schema__node__relationships", "schema__node__attributes"]
    AND r1.branch = default_branch AND r2.branch = default_branch
    AND r1.status = "active" AND r2.status = "active"
    AND r1.to IS NULL AND r2.to IS NULL
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
    RETURN peer AS schema_vertex
}
WITH default_branch, field, field_name, is_attribute, active_from, schema_vertex
ORDER BY active_from DESC
WITH default_branch, field_name, is_attribute, schema_vertex, collect(field) AS fields_reverse_chron
WHERE size(fields_reverse_chron) > 1
        """
        self.add_to_query(query)


class GetDuplicateSchemaFields(DuplicateSchemaFields):
    """
    Get the kind, field type, and field name for any duplicated attributes or relationships on a given schema
    on the default branch
    """

    name = "get_duplicate_schema_fields"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        await super().query_init(db=db, **kwargs)
        query = """
CALL (schema_vertex, default_branch) {
    MATCH (schema_vertex)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[r2:HAS_VALUE]->(name_value:AttributeValue)
    WHERE r1.branch = default_branch AND r2.branch = default_branch
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
    RETURN name_value.value AS schema_namespace
}
CALL (schema_vertex, default_branch) {
    MATCH (schema_vertex)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(name_value:AttributeValue)
    WHERE r1.branch = default_branch AND r2.branch = default_branch
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
    RETURN name_value.value AS schema_name
}
RETURN schema_namespace + schema_name AS schema_kind, schema_vertex.uuid AS schema_uuid, field_name, is_attribute
ORDER BY schema_kind ASC, is_attribute DESC, field_name ASC
        """
        self.return_labels = ["schema_kind", "schema_uuid", "field_name", "is_attribute"]
        self.add_to_query(query)

    def get_schema_field_details(self) -> list[SchemaFieldDetails]:
        schema_field_details: list[SchemaFieldDetails] = []
        for result in self.results:
            schema_kind = result.get_as_type(label="schema_kind", return_type=str)
            schema_uuid = result.get_as_type(label="schema_uuid", return_type=str)
            field_name = result.get_as_type(label="field_name", return_type=str)
            is_attribute = result.get_as_type(label="is_attribute", return_type=bool)
            schema_field_details.append(
                SchemaFieldDetails(
                    schema_kind=schema_kind,
                    schema_uuid=schema_uuid,
                    field_name=field_name,
                    field_type=SchemaFieldType.ATTRIBUTE if is_attribute else SchemaFieldType.RELATIONSHIP,
                )
            )
        return schema_field_details


class FixDuplicateSchemaFields(DuplicateSchemaFields):
    """
    Fix the duplicate schema fields by hard deleting the earlier duplicate(s)
    """

    name = "fix_duplicate_schema_fields"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        await super().query_init(db=db, **kwargs)
        query = """
WITH default_branch, tail(fields_reverse_chron) AS fields_to_delete
UNWIND fields_to_delete AS field_to_delete
CALL (field_to_delete, default_branch) {
    MATCH (field_to_delete)-[r:IS_PART_OF {branch: default_branch}]-()
    DELETE r
    WITH field_to_delete
    MATCH (field_to_delete)-[:IS_RELATED {branch: default_branch}]-(rel:Relationship)
    WITH DISTINCT field_to_delete, rel
    MATCH (rel)-[r {branch: default_branch}]-()
    DELETE r
    WITH field_to_delete, rel
    OPTIONAL MATCH (rel)
    WHERE NOT exists((rel)--())
    DELETE rel
    WITH DISTINCT field_to_delete
    MATCH (field_to_delete)-[:HAS_ATTRIBUTE {branch: default_branch}]->(attr:Attribute)
    MATCH (attr)-[r {branch: default_branch}]-()
    DELETE r
    WITH field_to_delete, attr
    OPTIONAL MATCH (attr)
    WHERE NOT exists((attr)--())
    DELETE attr
    WITH DISTINCT field_to_delete
    OPTIONAL MATCH (field_to_delete)
    WHERE NOT exists((field_to_delete)--())
    DELETE field_to_delete
}
        """
        self.add_to_query(query)


def display_duplicate_schema_fields(duplicate_schema_fields: list[SchemaFieldDetails]) -> None:
    console = Console()

    table = Table(title="Duplicate Schema Fields on Default Branch")

    table.add_column("Schema Kind")
    table.add_column("Schema UUID")
    table.add_column("Field Name")
    table.add_column("Field Type")

    for duplicate_schema_field in duplicate_schema_fields:
        table.add_row(
            duplicate_schema_field.schema_kind,
            duplicate_schema_field.schema_uuid,
            duplicate_schema_field.field_name,
            duplicate_schema_field.field_type.value,
        )

    console.print(table)


async def clean_duplicate_schema_fields(db: InfrahubDatabase, fix: bool = False) -> bool:
    """
    Identify any attributes or relationships that are duplicated in a schema on the default branch
    If fix is True, runs cypher queries to hard delete the earlier duplicate
    """

    duplicate_schema_fields_query = await GetDuplicateSchemaFields.init(db=db)
    await duplicate_schema_fields_query.execute(db=db)
    duplicate_schema_fields = duplicate_schema_fields_query.get_schema_field_details()

    if not duplicate_schema_fields:
        rprint(f"{SUCCESS_BADGE} No duplicate schema fields found")
        return True

    display_duplicate_schema_fields(duplicate_schema_fields)

    if not fix:
        rprint(f"{FAILED_BADGE} Use the --fix flag to fix the duplicate schema fields")
        return False

    fix_duplicate_schema_fields_query = await FixDuplicateSchemaFields.init(db=db)
    await fix_duplicate_schema_fields_query.execute(db=db)
    rprint(f"{SUCCESS_BADGE} Duplicate schema fields deleted from the default branch")
    return True



from __future__ import annotations

from typing import Any

from infrahub.core.constants import MetadataOptions
from infrahub.core.metadata.model import MetadataQueryOptions


def get_metadata_options_from_fields(fields: dict[str, Any]) -> MetadataOptions:
    """Convert a dict of requested metadata fields to MetadataOptions flags."""
    options = MetadataOptions.NONE
    if "created_at" in fields:
        options |= MetadataOptions.CREATED_AT
    if "created_by" in fields:
        options |= MetadataOptions.CREATED_BY
    if "updated_at" in fields or "_updated_at" in fields:
        options |= MetadataOptions.UPDATED_AT
    if "updated_by" in fields:
        options |= MetadataOptions.UPDATED_BY
    if "source" in fields or "_relation__source" in fields:
        options |= MetadataOptions.SOURCE
    if "owner" in fields or "_relation__owner" in fields:
        options |= MetadataOptions.OWNER
    if "is_protected" in fields:
        options |= MetadataOptions.IS_PROTECTED
    return options


def _extract_attribute_metadata_from_node_fields(node_fields: dict[str, Any]) -> MetadataOptions:
    """Extract attribute-level metadata options from nested node fields.

    This handles GraphQL query structures like:
        name {
            value
            updated_by { id }
            updated_at
        }

    Where `updated_by` and `updated_at` are attribute-level metadata.
    """
    attribute_metadata_options = MetadataOptions.NONE

    for field_properties_dict in node_fields.values():
        if not field_properties_dict or not isinstance(field_properties_dict, dict):
            continue

        # For attributes, metadata fields are at the same level as 'value'
        # Check if this looks like an attribute (has metadata fields directly)
        metadata_properties_dict = field_properties_dict

        # Extract metadata from attribute fields
        if "updated_at" in metadata_properties_dict or "_updated_at" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.UPDATED_AT
        if "updated_by" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.UPDATED_BY
        if "created_at" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.CREATED_AT
        if "created_by" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.CREATED_BY
        if "source" in metadata_properties_dict or "_relation__source" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.SOURCE
        if "owner" in metadata_properties_dict or "_relation__owner" in metadata_properties_dict:
            attribute_metadata_options |= MetadataOptions.OWNER

    return attribute_metadata_options


def build_metadata_query_options(
    node_metadata_fields: dict[str, Any] | None = None,
    relationship_metadata_fields: dict[str, Any] | None = None,
    node_fields: dict[str, Any] | None = None,
) -> MetadataQueryOptions:
    """Build MetadataQueryOptions from GraphQL extracted fields.

    Args:
        node_metadata_fields: Fields from the node_metadata section of a GraphQL query.
        relationship_metadata_fields: Fields from the relationship_metadata section.
        node_fields: The node fields dict to extract attribute-level metadata from.

    Returns:
        MetadataQueryOptions with appropriate flags set for each level.
    """
    node_level = get_metadata_options_from_fields(node_metadata_fields or {})
    relationship_level = get_metadata_options_from_fields(relationship_metadata_fields or {})
    attribute_level = _extract_attribute_metadata_from_node_fields(node_fields or {})

    return MetadataQueryOptions(
        node_level=node_level,
        attribute_level=attribute_level,
        relationship_level=relationship_level,
    )



from __future__ import annotations

from infrahub_sdk.exceptions import URLNotFoundError
from prefect import flow
from prefect.logging import get_run_logger

from infrahub.computed_attribute.jinja2 import InfrahubJinja2Template
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.registry import registry
from infrahub.events import BranchDeletedEvent
from infrahub.trigger.models import TriggerSetupReport, TriggerType
from infrahub.trigger.setup import setup_triggers_specific
from infrahub.workers.dependencies import get_client, get_component, get_database, get_workflow
from infrahub.workflows.catalogue import DISPLAY_LABELS_PROCESS_JINJA2, TRIGGER_UPDATE_DISPLAY_LABELS
from infrahub.workflows.utils import add_tags, wait_for_schema_to_converge

from .gather import gather_trigger_display_labels_jinja2
from .models import DisplayLabelJinja2GraphQL, DisplayLabelJinja2GraphQLResponse, DisplayLabelTriggerDefinition

UPDATE_DISPLAY_LABEL = """
mutation UpdateDisplayLabel(
    $id: String!,
    $kind: String!,
    $value: String!,
    $context_account_id: String!
  ) {
  InfrahubUpdateDisplayLabel(
    context: {account: {id: $context_account_id}},
    data: {id: $id, value: $value, kind: $kind}
  ) {
    ok
  }
}
"""


@flow(
    name="display-label-jinja2-update-value",
    flow_run_name="Update value for display_label on {node_kind}",
)
async def display_label_jinja2_update_value(
    branch_name: str,
    obj: DisplayLabelJinja2GraphQLResponse,
    node_kind: str,
    template: InfrahubJinja2Template,
    context: InfrahubContext,
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name], nodes=[obj.node_id], db_change=True)

    value = await template.render(variables=obj.variables)
    if value == obj.display_label_value:
        log.debug(f"Ignoring to update {obj} with existing value on display_label={value}")
        return

    try:
        await client.execute_graphql(
            query=UPDATE_DISPLAY_LABEL,
            variables={
                "id": obj.node_id,
                "kind": node_kind,
                "value": value,
                "context_account_id": context.account.account_id,
            },
            branch_name=branch_name,
        )
        log.info(f"Updating {node_kind}.display_label='{value}' ({obj.node_id})")
    except URLNotFoundError:
        log.warning(
            f"Updating {node_kind}.display_label='{value}' ({obj.node_id}) failed for branch {branch_name} (branch not found)"
        )


@flow(
    name="display-label-process-jinja2",
    flow_run_name="Process display_labels for {target_kind}",
)
async def process_display_label(
    branch_name: str,
    node_kind: str,
    object_id: str,
    target_kind: str,
    context: InfrahubContext,
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name])

    target_schema = branch_name if branch_name in registry.get_altered_schema_branches() else registry.default_branch
    schema_branch = registry.schema.get_schema_branch(name=target_schema)
    node_schema = schema_branch.get_node(name=target_kind, duplicate=False)

    if node_kind == target_kind:
        display_label_template = schema_branch.display_labels.get_template_node(kind=node_kind)
    else:
        display_label_template = schema_branch.display_labels.get_related_template(
            related_kind=node_kind, target_kind=target_kind
        )

    jinja_template = InfrahubJinja2Template(template=display_label_template.template)
    variables = jinja_template.get_variables()
    display_label_graphql = DisplayLabelJinja2GraphQL(
        node_schema=node_schema, variables=variables, filter_key=display_label_template.filter_key
    )

    query = display_label_graphql.render_graphql_query(filter_id=object_id)
    response = await client.execute_graphql(query=query, branch_name=branch_name)
    update_candidates = display_label_graphql.parse_response(response=response)

    if not update_candidates:
        log.debug("No nodes found that requires updates")
        return

    batch = await client.create_batch()
    for node in update_candidates:
        batch.add(
            task=display_label_jinja2_update_value,
            branch_name=branch_name,
            obj=node,
            node_kind=node_schema.kind,
            template=jinja_template,
            context=context,
        )

    _ = [response async for _, response in batch.execute()]


@flow(name="display-labels-setup-jinja2", flow_run_name="Setup display labels in task-manager")
async def display_labels_setup_jinja2(
    context: InfrahubContext, branch_name: str | None = None, event_name: str | None = None
) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        if branch_name:
            await add_tags(branches=[branch_name])
            component = await get_component()
            await wait_for_schema_to_converge(branch_name=branch_name, component=component, db=db, log=log)

        report: TriggerSetupReport = await setup_triggers_specific(
            gatherer=gather_trigger_display_labels_jinja2, trigger_type=TriggerType.DISPLAY_LABEL_JINJA2
        )  # type: ignore[misc]

        # Configure all DisplayLabelTriggerDefinitions in Prefect
        all_triggers = report.triggers_with_type(trigger_type=DisplayLabelTriggerDefinition)
        direct_target_triggers = [
            display_report
            for display_report in report.modified_triggers_with_type(trigger_type=DisplayLabelTriggerDefinition)
            if display_report.target_kind
        ]

        for display_report in direct_target_triggers:
            if event_name != BranchDeletedEvent.event_name and display_report.branch == branch_name:
                if branch_name != registry.default_branch:
                    default_branch_triggers = [
                        trigger
                        for trigger in all_triggers
                        if trigger.branch == registry.default_branch
                        and trigger.target_kind == display_report.target_kind
                    ]
                    if (
                        default_branch_triggers
                        and len(default_branch_triggers) == 1
                        and default_branch_triggers[0].template_hash == display_report.template_hash
                    ):
                        log.debug(
                            f"Skipping display label updates for {display_report.target_kind} [{branch_name}], schema is identical to default branch"
                        )
                        continue

                await get_workflow().submit_workflow(
                    workflow=TRIGGER_UPDATE_DISPLAY_LABELS,
                    context=context,
                    parameters={
                        "branch_name": display_report.branch,
                        "kind": display_report.target_kind,
                    },
                )

        log.info(f"{report.in_use_count} Display labels for Jinja2 automation configuration completed")


@flow(
    name="trigger-update-display-labels",
    flow_run_name="Trigger updates for display labels for {kind}",
)
async def trigger_update_display_labels(
    branch_name: str,
    kind: str,
    context: InfrahubContext,
) -> None:
    await add_tags(branches=[branch_name])

    client = get_client()

    # NOTE we only need the id of the nodes, this query will still query for the HFID
    node_schema = registry.schema.get_node_schema(name=kind, branch=branch_name)
    nodes = await client.all(
        kind=kind,
        branch=branch_name,
        exclude=node_schema.attribute_names + node_schema.relationship_names,
        populate_store=False,
    )

    for node in nodes:
        await get_workflow().submit_workflow(
            workflow=DISPLAY_LABELS_PROCESS_JINJA2,
            context=context,
            parameters={
                "branch_name": branch_name,
                "node_kind": kind,
                "target_kind": kind,
                "object_id": node.id,
                "context": context,
            },
        )



from __future__ import annotations

import contextlib
import io
from typing import TYPE_CHECKING, Any, BinaryIO

import botocore.exceptions
import fastapi_storages
from typing_extensions import Self

from infrahub.exceptions import NodeNotFoundError

if TYPE_CHECKING:
    from infrahub.config import StorageSettings


class InfrahubS3ObjectStorage(fastapi_storages.S3Storage):
    def __init__(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        super().__init__()

    def open(self, name: str) -> BinaryIO:
        f = io.BytesIO()
        self._bucket.download_fileobj(name, f)
        f.flush()
        f.seek(0)
        return f  # type: ignore

    def delete(self, name: str) -> None:
        self._bucket.Object(name).delete()


fastapi_storages.InfrahubS3ObjectStorage = InfrahubS3ObjectStorage


class InfrahubObjectStorage:
    _settings: StorageSettings
    _storage: fastapi_storages.base.BaseStorage

    def __init__(self, settings: StorageSettings) -> None:
        self._settings = settings

        driver = getattr(fastapi_storages, self._settings.driver.name)

        driver_settings = getattr(self._settings, self._settings.driver.value.lower())
        self._storage = driver(**driver_settings.model_dump(by_alias=True))

    @classmethod
    async def init(cls, settings: StorageSettings) -> Self:
        return cls(settings)

    def store(self, identifier: str, content: BinaryIO) -> None:
        self._storage.write(content, identifier)

    def retrieve(self, identifier: str) -> str:
        try:
            with self._storage.open(identifier) as f:
                return f.read().decode()
        except (FileNotFoundError, botocore.exceptions.ClientError) as err:
            raise NodeNotFoundError(node_type="StorageObject", identifier=identifier) from err

    def retrieve_binary(self, identifier: str) -> bytes:
        try:
            with self._storage.open(identifier) as f:
                return f.read()
        except (FileNotFoundError, botocore.exceptions.ClientError) as err:
            raise NodeNotFoundError(node_type="StorageObject", identifier=identifier) from err

    def delete(self, identifier: str) -> None:
        """Delete a file from storage.

        Args:
            identifier: The storage identifier of the file to delete.

        Note:
            Silently ignores if the file does not exist.
        """
        if isinstance(self._storage, fastapi_storages.FileSystemStorage):
            (self._storage._path / identifier).unlink(missing_ok=True)
        else:
            with contextlib.suppress(FileNotFoundError, botocore.exceptions.ClientError):
                self._storage.delete(identifier)



from __future__ import annotations


class NodeProperty:
    """Represents a property of a node, typically used for metadata like display labels."""

    def __init__(self, data: dict | str) -> None:
        """
        Args:
            data (Union[dict, str]): Data representing the node property.
        """
        self.id = None
        self.display_label = None
        self.typename = None

        if isinstance(data, str):
            self.id = data
        elif isinstance(data, dict):
            self.id = data.get("id", None)
            self.display_label = data.get("display_label", None)
            self.typename = data.get("__typename", None)

    def __repr__(self) -> str:
        return f"NodeProperty({{'id': {self.id!r}, 'display_label': {self.display_label!r}, '__typename': {self.typename!r}}})"

    def _generate_input_data(self) -> str | None:
        return self.id



from __future__ import annotations

from infrahub.constants.database import EntityType, IndexType
from infrahub.core.query import QueryType

from .index import IndexInfo, IndexItem, IndexManagerBase


class IndexRelNeo4j(IndexItem):
    @property
    def _index_name(self) -> str:
        return f"rel_{self.type.value.lower()}_{self.name}_{'_'.join(self.properties)}"

    def get_add_query(self) -> str:
        properties_str = ", ".join([f"r.{prop}" for prop in self.properties])
        return (
            f"CREATE {self.type.value.upper()} INDEX {self._index_name} IF NOT EXISTS "
            f"FOR ()-[r:{self.label}]-() ON ({properties_str})"
        )

    def get_drop_query(self) -> str:
        return f"DROP INDEX {self._index_name} IF EXISTS"


class IndexNodeNeo4j(IndexItem):
    @property
    def _index_name(self) -> str:
        return f"node_{self.type.value.lower()}_{self.name}_{'_'.join(self.properties)}"

    def get_add_query(self) -> str:
        properties_str = ", ".join([f"n.{prop}" for prop in self.properties])
        return (
            f"CREATE {self.type.value.upper()} INDEX {self._index_name} IF NOT EXISTS "
            f"FOR (n:{self.label}) ON ({properties_str})"
        )

    def get_drop_query(self) -> str:
        return f"DROP INDEX {self._index_name} IF EXISTS"


class IndexManagerNeo4j(IndexManagerBase):
    def init(self, nodes: list[IndexItem], rels: list[IndexItem]) -> None:
        self.nodes = [IndexNodeNeo4j(**item.model_dump()) for item in nodes]
        self.rels = [IndexRelNeo4j(**item.model_dump()) for item in rels]
        self.initialized = True

    async def list(self) -> list[IndexInfo]:
        query = "SHOW INDEXES"
        records = await self.db.execute_query(query=query, params={}, name="index_show", type=QueryType.READ)
        results = []
        for record in records:
            if not record["labelsOrTypes"]:
                continue
            results.append(
                IndexInfo(
                    name=record["name"],
                    label=", ".join(record["labelsOrTypes"]),
                    properties=record["properties"],
                    type=IndexType(str(record["type"]).lower()),
                    entity_type=EntityType(str(record["entityType"]).lower()),
                )
            )

        return results



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub.core import registry
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.migrations.graph import MIGRATIONS

from .exceptions import MigrationFailureError
from .shared import MigrationInput, MigrationRequiringRebase

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class MigrationRunner:
    def __init__(self, branch: Branch) -> None:
        if branch.name in (registry.default_branch, GLOBAL_BRANCH_NAME):
            raise ValueError("MigrationRunner cannot be used to apply migration on default branches")

        self.branch = branch
        self.applicable_migrations = self._get_applicable_migrations()

    def _get_applicable_migrations(self) -> Sequence[MigrationRequiringRebase]:
        applicable_migrations = []
        for migration_class in [m for m in MIGRATIONS if issubclass(m, MigrationRequiringRebase)]:
            migration = migration_class.init()
            if self.branch.graph_version and self.branch.graph_version > migration.minimum_version:
                continue
            applicable_migrations.append(migration)

        return applicable_migrations

    def has_migrations(self) -> bool:
        return bool(self.applicable_migrations)

    async def run(self, db: InfrahubDatabase, at: Timestamp) -> None:
        if not self.has_migrations():
            return

        for migration in self.applicable_migrations:
            execution_result = await migration.execute_against_branch(
                migration_input=MigrationInput(db=db, at=at), branch=self.branch
            )
            validation_result = None

            if execution_result.success:
                validation_result = await migration.validate_migration(db=db)

            if not execution_result.success or (validation_result and not validation_result.success):
                if execution_result.errors:
                    raise MigrationFailureError(errors=execution_result.errors)

                if validation_result and not validation_result.success:
                    raise MigrationFailureError(errors=validation_result.errors)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from fastapi import APIRouter, Depends, Query, Request
from pydantic import (
    BaseModel,
    Field,
    computed_field,
    create_model,
    model_validator,
)
from starlette.responses import JSONResponse

from infrahub import lock
from infrahub.api.dependencies import get_branch_dep, get_context, get_current_user, get_db, get_permission_manager
from infrahub.api.exceptions import SchemaNotValidError
from infrahub.branch.status_checker import BranchStatusChecker
from infrahub.core import registry
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, PermissionDecision
from infrahub.core.models import (  # noqa: TC001
    SchemaBranchHash,
    SchemaDiff,
    SchemaUpdateConstraintInfo,
    SchemaUpdateValidationResult,
)
from infrahub.core.schema import (
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    SchemaRoot,
    SchemaWarning,
    TemplateSchema,
)
from infrahub.core.schema.constants import SchemaNamespace  # noqa: TC001
from infrahub.core.schema.update_coordinator import MigrationExecutor, SchemaUpdateCoordinator
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.models.validate_migration import (
    SchemaValidateMigrationData,
    SchemaValidatorPathResponseData,
)
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.events import EventMeta
from infrahub.events.schema_action import SchemaUpdatedEvent
from infrahub.exceptions import BranchStatusError, ValidationError
from infrahub.log import get_log_data, get_logger
from infrahub.permissions import define_global_permission_from_branch
from infrahub.types import ATTRIBUTE_PYTHON_TYPES
from infrahub.worker import WORKER_IDENTITY
from infrahub.workflows.catalogue import SCHEMA_VALIDATE_MIGRATION

if TYPE_CHECKING:
    from typing_extensions import Self

    from infrahub.auth import AccountSession
    from infrahub.context import InfrahubContext
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.permissions import PermissionManager
    from infrahub.services import InfrahubServices


log = get_logger()
router = APIRouter(prefix="/schema")


class APISchemaMixin:
    @classmethod
    def from_schema(cls, schema: MainSchemaTypes) -> Self:
        data = schema.model_dump()
        data["relationships"] = [
            relationship.model_dump() for relationship in schema.relationships if not relationship.internal_peer
        ]
        data["hash"] = schema.get_hash()
        return cls(**data)

    @model_validator(mode="before")
    @classmethod
    def set_kind(cls, values: Any) -> Any:
        if isinstance(values, dict):
            values["kind"] = f"{values['namespace']}{values['name']}"
        return values


class APINodeSchema(NodeSchema, APISchemaMixin):
    api_kind: str | None = Field(default=None, alias="kind", validate_default=True)
    hash: str


class APIGenericSchema(GenericSchema, APISchemaMixin):
    api_kind: str | None = Field(default=None, alias="kind", validate_default=True)
    hash: str


class APIProfileSchema(ProfileSchema, APISchemaMixin):
    api_kind: str | None = Field(default=None, alias="kind", validate_default=True)
    hash: str


class APITemplateSchema(TemplateSchema, APISchemaMixin):
    api_kind: str | None = Field(default=None, alias="kind", validate_default=True)
    hash: str


class SchemaReadAPI(BaseModel):
    main: str = Field(description="Main hash for the entire schema")
    nodes: list[APINodeSchema] = Field(default_factory=list)
    generics: list[APIGenericSchema] = Field(default_factory=list)
    profiles: list[APIProfileSchema] = Field(default_factory=list)
    templates: list[APITemplateSchema] = Field(default_factory=list)
    namespaces: list[SchemaNamespace] = Field(default_factory=list)


class SchemaLoadAPI(SchemaRoot):
    version: str


class SchemasLoadAPI(BaseModel):
    schemas: list[SchemaLoadAPI]


class JSONSchema(BaseModel):
    title: str | None = Field(None, description="Title of the schema")
    description: str | None = Field(None, description="Description of the schema")
    type: str = Field(..., description="Type of the schema element (e.g., 'object', 'array', 'string')")
    properties: dict[str, Any] | None = Field(None, description="Properties of the object if type is 'object'")
    items: dict[str, Any] | list[dict[str, Any]] | None = Field(
        None, description="Items of the array if type is 'array'"
    )
    required: list[str] | None = Field(None, description="List of required properties if type is 'object'")
    schema_spec: str | None = Field(None, alias="$schema", description="Schema version identifier")
    additional_properties: bool | dict[str, Any] | None = Field(
        None, description="Specifies whether additional properties are allowed", alias="additionalProperties"
    )


class SchemaUpdate(BaseModel):
    hash: str = Field(..., description="The new hash for the entire schema")
    previous_hash: str = Field(..., description="The previous hash for the entire schema")
    diff: SchemaDiff = Field(..., description="The modifications to the schema")
    warnings: list[SchemaWarning] = Field(
        default_factory=list, description="Warnings encountered while loading the schema"
    )

    @computed_field
    def schema_updated(self) -> bool:
        """Indicates if the loading of the schema changed the existing schema"""
        return self.hash != self.previous_hash


def _merge_candidate_schemas(schemas: Sequence[SchemaRoot]) -> SchemaRoot:
    """Merge multiple schemas into one suitable to be loaded."""
    if not schemas:
        raise ValueError("Cannot merge an empty list of schemas")

    merged = schemas[0]
    for schema in schemas[1:]:
        merged = merged.merge(schema=schema)

    return merged


def evaluate_candidate_schemas(
    branch_schema: SchemaBranch, schemas_to_evaluate: SchemasLoadAPI
) -> tuple[SchemaBranch, SchemaUpdateValidationResult]:
    candidate_schema = branch_schema.duplicate()
    try:
        schema = _merge_candidate_schemas(schemas=schemas_to_evaluate.schemas)

        candidate_schema.load_schema(schema=schema)
        candidate_schema.process()

        schema_diff = branch_schema.diff(other=candidate_schema)
        candidate_schema.validate_node_deletions(diff=schema_diff)
    except ValueError as exc:
        raise SchemaNotValidError(message=str(exc)) from exc

    result = branch_schema.validate_update(other=candidate_schema, diff=schema_diff)

    if result.errors:
        raise SchemaNotValidError(message=", ".join([error.to_string() for error in result.errors]))

    return candidate_schema, result


@router.get("")
async def get_schema(
    branch: Branch = Depends(get_branch_dep),
    namespaces: list[str] | None = Query(default=None),
    _: AccountSession = Depends(get_current_user),
) -> SchemaReadAPI:
    log.debug("schema_request", branch=branch.name)
    schema_branch = registry.schema.get_schema_branch(name=branch.name)
    all_schemas = schema_branch.get_schemas_for_namespaces(namespaces=namespaces)

    return SchemaReadAPI(
        main=registry.schema.get_schema_branch(name=branch.name).get_hash(),
        nodes=[
            APINodeSchema.from_schema(value)
            for value in all_schemas
            if isinstance(value, NodeSchema) and value.namespace != "Internal"
        ],
        generics=[
            APIGenericSchema.from_schema(value)
            for value in all_schemas
            if isinstance(value, GenericSchema) and value.namespace != "Internal"
        ],
        profiles=[
            APIProfileSchema.from_schema(value)
            for value in all_schemas
            if isinstance(value, ProfileSchema) and value.namespace != "Internal"
        ],
        templates=[
            APITemplateSchema.from_schema(value)
            for value in all_schemas
            if isinstance(value, TemplateSchema) and value.namespace != "Internal"
        ],
        namespaces=schema_branch.get_namespaces(),
    )


@router.get("/summary")
async def get_schema_summary(
    branch: Branch = Depends(get_branch_dep), _: AccountSession = Depends(get_current_user)
) -> SchemaBranchHash:
    log.debug("schema_summary_request", branch=branch.name)
    schema_branch = registry.schema.get_schema_branch(name=branch.name)
    return schema_branch.get_hash_full()


@router.get("/{schema_kind}")
async def get_schema_by_kind(
    schema_kind: str, branch: Branch = Depends(get_branch_dep), _: AccountSession = Depends(get_current_user)
) -> APIProfileSchema | APINodeSchema | APIGenericSchema | APITemplateSchema:
    log.debug("schema_kind_request", branch=branch.name)

    schema = registry.schema.get(name=schema_kind, branch=branch, duplicate=False)

    api_schema: dict[str, type[APIProfileSchema | APINodeSchema | APIGenericSchema | APITemplateSchema]] = {
        "profile": APIProfileSchema,
        "node": APINodeSchema,
        "generic": APIGenericSchema,
        "template": APITemplateSchema,
    }
    key = ""

    if isinstance(schema, ProfileSchema):
        key = "profile"
    if isinstance(schema, NodeSchema):
        key = "node"
    if isinstance(schema, GenericSchema):
        key = "generic"
    if isinstance(schema, TemplateSchema):
        key = "template"

    return api_schema[key].from_schema(schema=schema)


@router.get("/json_schema/{schema_kind}")
async def get_json_schema_by_kind(
    schema_kind: str, branch: Branch = Depends(get_branch_dep), _: AccountSession = Depends(get_current_user)
) -> JSONSchema:
    log.debug("json_schema_kind_request", branch=branch.name)

    fields: dict[str, Any] = {}

    schema = registry.schema.get(name=schema_kind, branch=branch)

    for attr in schema.attributes:
        field_type = ATTRIBUTE_PYTHON_TYPES[attr.kind]

        default_value = attr.default_value if attr.optional else ...
        field_info = Field(default=default_value, description=attr.description)
        if attr.enum or attr.kind == "Dropdown":
            extras: dict[str, Any]
            if attr.kind == "Dropdown" and attr.choices:
                extras = {"enum": [choice.name for choice in attr.choices]}
            else:
                extras = {"enum": attr.enum}
            field_info = Field(default=default_value, description=attr.description, json_schema_extra=extras)
        fields[attr.name] = (field_type, field_info)

    # Use Pydantic's create_model to dynamically create the class, ignore types because fields are Any, and mypy hates that
    json_schema = create_model(schema.name, **fields).model_json_schema()

    json_schema["description"] = schema.description
    json_schema["$schema"] = "http://json-schema.org/draft-07/schema#"

    return json_schema


async def _validate_migrations(
    branch: Branch,
    candidate_schema: SchemaBranch,
    constraints: list[SchemaUpdateConstraintInfo],
    service: InfrahubServices,
    context: InfrahubContext,
) -> None:
    validate_migration_data = SchemaValidateMigrationData(
        branch=branch,
        schema_branch=candidate_schema,
        constraints=constraints,
    )
    responses = await service.workflow.execute_workflow(
        workflow=SCHEMA_VALIDATE_MIGRATION,
        context=context,
        expected_return=list[SchemaValidatorPathResponseData],
        parameters={"message": validate_migration_data},
    )
    error_messages = [violation.message for response in responses for violation in response.violations]
    if error_messages:
        raise SchemaNotValidError(",\n".join(error_messages))


@router.post("/load")
async def load_schema(
    request: Request,
    schemas: SchemasLoadAPI,
    db: InfrahubDatabase = Depends(get_db),
    branch: Branch = Depends(get_branch_dep),
    account_session: AccountSession = Depends(get_current_user),
    permission_manager: PermissionManager = Depends(get_permission_manager),
    context: InfrahubContext = Depends(get_context),
) -> SchemaUpdate:
    try:
        BranchStatusChecker().check(branch=branch)
    except BranchStatusError as err:
        raise ValidationError(input_value=str(err)) from err

    permission_manager.raise_for_permission(
        permission=define_global_permission_from_branch(
            permission=GlobalPermissions.MANAGE_SCHEMA, branch_name=branch.name
        )
    )

    if branch.name in (GLOBAL_BRANCH_NAME, registry.default_branch):
        permission_manager.raise_for_permission(
            permission=GlobalPermission(
                action=GlobalPermissions.EDIT_DEFAULT_BRANCH.value, decision=PermissionDecision.ALLOW_DEFAULT.value
            ),
        )

    service: InfrahubServices = request.app.state.service
    log.info("schema_load_request", branch=branch.name)

    errors: list[str] = []
    warnings: list[SchemaWarning] = []
    for schema in schemas.schemas:
        errors += schema.validate_namespaces()
        errors += schema.validate_reserved_suffixes()
        warnings += schema.gather_warnings()

    if errors:
        raise SchemaNotValidError(message=", ".join(errors))

    async with lock.registry.global_schema_lock():
        branch_schema = registry.schema.get_schema_branch(name=branch.name)
        original_hash = branch_schema.get_hash()

        candidate_schema, result = evaluate_candidate_schemas(branch_schema=branch_schema, schemas_to_evaluate=schemas)

        if not result.diff.all:
            return SchemaUpdate(hash=original_hash, previous_hash=original_hash, diff=result.diff)

        # ----------------------------------------------------------
        # Validate if the new schema is valid with the content of the database
        # ----------------------------------------------------------
        await _validate_migrations(
            branch=branch,
            candidate_schema=candidate_schema,
            constraints=result.constraints,
            service=service,
            context=context,
        )

        origin_schema = branch_schema.duplicate()

        log.info("Schema has diff, will need to be updated", diff=result.diff.all, branch=branch.name)

        coordinator = SchemaUpdateCoordinator(
            db=db,
            branch=branch,
            schema_manager=registry.schema,
            origin_schema=origin_schema,
            workflow=service.workflow,
            context=context,
            migration_executor=MigrationExecutor.WORKFLOW,
        )

        updated_hash = await coordinator.execute(
            candidate_schema=candidate_schema,
            at=Timestamp(),
            diff=result.diff,
            migrations=result.migrations,
            limit=result.diff.all,
            update_db=True,
            user_id=account_session.account_id,
        )

    await service.component.refresh_schema_hash(branches=[branch.name])

    log_data = get_log_data()
    request_id = log_data.get("request_id", "")
    event = SchemaUpdatedEvent(
        branch_name=branch.name,
        schema_hash=branch.active_schema_hash.main,
        meta=EventMeta(
            initiator_id=WORKER_IDENTITY,
            request_id=request_id,
            account_id=account_session.account_id,
            branch=branch,
            context=context,
        ),
    )
    await service.event.send(event=event)

    return SchemaUpdate(hash=updated_hash, previous_hash=original_hash, diff=result.diff, warnings=warnings)


@router.post("/check")
async def check_schema(
    request: Request,
    schemas: SchemasLoadAPI,
    branch: Branch = Depends(get_branch_dep),
    context: InfrahubContext = Depends(get_context),
    _: AccountSession = Depends(get_current_user),
) -> JSONResponse:
    service: InfrahubServices = request.app.state.service
    log.info("schema_check_request", branch=branch.name)

    errors: list[str] = []
    warnings: list[SchemaWarning] = []
    for schema in schemas.schemas:
        errors += schema.validate_namespaces()
        errors += schema.validate_reserved_suffixes()
        warnings += schema.gather_warnings()

    if errors:
        raise SchemaNotValidError(message=", ".join(errors))

    branch_schema = registry.schema.get_schema_branch(name=branch.name)

    candidate_schema, result = evaluate_candidate_schemas(branch_schema=branch_schema, schemas_to_evaluate=schemas)

    # ----------------------------------------------------------
    # Validate if the new schema is valid with the content of the database
    # ----------------------------------------------------------
    validate_migration_data = SchemaValidateMigrationData(
        branch=branch,
        schema_branch=candidate_schema,
        constraints=result.constraints,
    )
    responses = await service.workflow.execute_workflow(
        workflow=SCHEMA_VALIDATE_MIGRATION,
        context=context,
        expected_return=list[SchemaValidatorPathResponseData],
        parameters={"message": validate_migration_data},
    )
    error_messages = [violation.message for response in responses for violation in response.violations]
    if error_messages:
        raise SchemaNotValidError(message=",\n".join(error_messages))

    return JSONResponse(
        status_code=202,
        content={
            "diff": result.diff.model_dump(),
            "warnings": [warning.model_dump(mode="json") for warning in warnings],
        },
    )



from __future__ import annotations

import contextlib
from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import ujson
from rich.progress import Progress, TaskID

from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.initialization import get_root_node
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)
from infrahub.core.node.node_property_attribute import DisplayLabel, HumanFriendlyIdentifier
from infrahub.core.query import Query, QueryType
from infrahub.core.query.node import AttributeFromDB
from infrahub.core.schema import GenericSchema, SchemaNotFoundError

from .load_schema_branch import get_or_load_schema_branch
from .m047_backfill_or_null_display_label import UpdateAttributeValuesQuery

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.schema import NodeSchema, ProfileSchema, TemplateSchema
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


console = get_migration_console()


@dataclass
class BadNodeInfo:
    node_uuid: str
    node_kind: str
    bad_attrs: list[str] = field(default_factory=list)


class FindNodesWithBadValues(Query):
    """Find nodes where display_label or human_friendly_id attribute values contain 'None' or 'null'.

    Starts from AttributeValue vertices (smaller scan) and traces back to nodes,
    rather than scanning all Node vertices (expensive on large databases).
    """

    name = "find_nodes_with_bad_hfid_display_label"
    type = QueryType.READ
    insert_return = False

    def __init__(
        self, value_branch_names: list[str], structural_branch_names: list[str] | None = None, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.value_branch_names = value_branch_names
        # For default/global branches, structural edges are on the same branches as values.
        # For user branches, structural edges (HAS_ATTRIBUTE, IS_PART_OF) may be on the default/global branch.
        self.structural_branch_names = structural_branch_names or value_branch_names

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["value_branch_names"] = self.value_branch_names
        self.params["structural_branch_names"] = self.structural_branch_names
        self.return_labels = ["node_uuid", "node_kind", "bad_attrs"]

        query = """
// Find candidate bad values from both indexed and non-indexed attribute values
CALL () {
    // Start with the indexed vertices
    MATCH (av:AttributeValueIndexed)
    WHERE av.value CONTAINS "None" OR av.value CONTAINS "null"
    RETURN av
    UNION
    // There should be fewer non-indexed vertices, so this scan will hopefully not be too big
    MATCH (av:AttributeValue&!AttributeValueIndexed)
    WHERE av.value CONTAINS "None" OR av.value CONTAINS "null"
    RETURN av
}
// Trace back to Attribute and filter by name; HAS_VALUE must be on the target branch
WITH av
MATCH (attr:Attribute)-[hv:HAS_VALUE]->(av)
WHERE attr.name IN ["display_label", "human_friendly_id"]
AND hv.branch IN $value_branch_names
AND hv.to IS NULL
AND hv.status = "active"
WITH attr, av
// Trace back to Node; HAS_ATTRIBUTE may be on a parent branch
MATCH (n:Node)-[ha:HAS_ATTRIBUTE]->(attr)
WHERE ha.branch IN $structural_branch_names
AND ha.to IS NULL
AND ha.status = "active"
WITH DISTINCT n, collect(DISTINCT attr.name) AS bad_attrs
// Verify node is active; IS_PART_OF may be on a parent branch
MATCH (n)-[e:IS_PART_OF]->(:Root)
WHERE e.branch IN $structural_branch_names
AND e.to IS NULL
AND e.status = "active"
RETURN n.uuid AS node_uuid, n.kind AS node_kind, bad_attrs
        """
        self.add_to_query(query)

    def get_bad_nodes(self) -> list[BadNodeInfo]:
        """Return list of BadNodeInfo for nodes with bad attribute values."""
        results: list[BadNodeInfo] = []
        for result in self.get_results():
            results.append(
                BadNodeInfo(
                    node_uuid=str(result.get("node_uuid")),
                    node_kind=str(result.get("node_kind")),
                    bad_attrs=[str(a) for a in result.get("bad_attrs")],
                )
            )
        return results


class Migration059(MigrationRequiringRebase):
    """Fix display_label and human_friendly_id attributes that contain 'None' or 'null' string values.

    These bad values were introduced by earlier migrations (m044-m047) when relationship peers
    were missing and the path value was stringified as 'None'.
    """

    name: str = "059_fix_hfid_display_label_nulls"
    minimum_version: int = 58
    update_batch_size: int = 100

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def _find_bad_nodes(
        self,
        db: InfrahubDatabase,
        value_branch_names: list[str],
        structural_branch_names: list[str] | None = None,
    ) -> list[BadNodeInfo]:
        find_query = await FindNodesWithBadValues.init(
            db=db, value_branch_names=value_branch_names, structural_branch_names=structural_branch_names
        )
        await find_query.execute(db=db)
        return find_query.get_bad_nodes()

    async def _compute_display_label(
        self,
        db: InfrahubDatabase,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        node: Node,
    ) -> str | None:
        """Compute the correct display_label value for a node."""
        if schema.display_label:
            dl = DisplayLabel(node_schema=schema, template=schema.display_label)
            # in case computing fails, will log an error later
            with contextlib.suppress(ValueError):
                await dl.compute(db=db, node=node)
            # NodePropertyAttribute doesn't expose a public getter that avoids needing a Node+Timestamp,
            # so accessing _value directly is the pragmatic choice in migration context.
            value = dl._value
            if value is not None:
                return value.value if isinstance(value, AttributeFromDB) else value
            console.print(
                f"[red]ERROR: Failed to compute display_label for node {node.id} ({schema.kind}) "
                f"using template {schema.display_label!r}[/red]"
            )
        # display_labels (plural) is deprecated but existing schemas in production may still use it
        elif schema.display_labels:
            parts = []
            for path in schema.display_labels:
                path_value = await node.get_path_value(db=db, path=path)
                if path_value is not None:
                    parts.append(str(path_value))
            if parts:
                return " ".join(parts)
            console.print(
                f"[red]ERROR: Failed to compute display_label for node {node.id} ({schema.kind}) "
                f"using display_labels paths {schema.display_labels!r}[/red]"
            )
        return None

    async def _compute_hfid(
        self,
        db: InfrahubDatabase,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        node: Node,
    ) -> str | None:
        """Compute the correct human_friendly_id value for a node, JSON-encoded."""
        if not schema.human_friendly_id:
            return None
        hfid = HumanFriendlyIdentifier(node_schema=schema, template=schema.human_friendly_id)
        # in case computing fails, will log an error later
        with contextlib.suppress(ValueError):
            await hfid.compute(db=db, node=node)
        hfid_value = hfid._value
        if hfid_value is not None:
            raw_list = hfid_value.value if isinstance(hfid_value, AttributeFromDB) else hfid_value
            return ujson.dumps(raw_list)
        console.print(
            f"[red]ERROR: Failed to compute human_friendly_id for node {node.id} ({schema.kind}) "
            f"using paths {schema.human_friendly_id!r}[/red]"
        )
        return None

    async def _compute_values_for_batch(
        self,
        db: InfrahubDatabase,
        load_branch: Branch,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        batch: list[BadNodeInfo],
        errors: list[str],
    ) -> tuple[dict[str, str], dict[str, str]]:
        """Load nodes and compute corrected display_label/HFID values for a batch.

        Returns (dl_values, hfid_values) dicts mapping node UUID to corrected value.
        """
        kind = schema.kind
        batch_uuids = [info.node_uuid for info in batch]
        bad_attrs_by_uuid = {info.node_uuid: info.bad_attrs for info in batch}

        loaded_nodes = await NodeManager.get_many(db=db, ids=batch_uuids, branch=load_branch)

        dl_values: dict[str, str] = {}
        hfid_values: dict[str, str] = {}

        for node_uuid, node in loaded_nodes.items():
            bad_attrs = bad_attrs_by_uuid.get(node_uuid, [])

            if "display_label" in bad_attrs and (schema.display_label or schema.display_labels):
                try:
                    value = await self._compute_display_label(db=db, schema=schema, node=node)
                    if value is not None:
                        dl_values[node_uuid] = value
                except Exception as exc:
                    console.print(f"  Skipping display_label for {node_uuid} ({kind}): {exc}")
                    errors.append(f"display_label compute failed for {node_uuid} ({kind}): {exc}")

            if "human_friendly_id" in bad_attrs and schema.human_friendly_id:
                try:
                    value = await self._compute_hfid(db=db, schema=schema, node=node)
                    if value is not None:
                        hfid_values[node_uuid] = value
                except Exception as exc:
                    console.print(f"  Skipping human_friendly_id for {node_uuid} ({kind}): {exc}")
                    errors.append(f"human_friendly_id compute failed for {node_uuid} ({kind}): {exc}")

        return dl_values, hfid_values

    async def _recompute_and_update(
        self,
        db: InfrahubDatabase,
        write_branch: Branch,
        load_branch: Branch,
        schema_branch: SchemaBranch,
        bad_nodes: list[BadNodeInfo],
        at: Timestamp,
        progress: Progress | None = None,
        progress_task: TaskID | None = None,
    ) -> MigrationResult:
        errors: list[str] = []

        # Group by kind for schema lookup and batching
        nodes_by_kind: dict[str, list[BadNodeInfo]] = defaultdict(list)
        for bad_node in bad_nodes:
            nodes_by_kind[bad_node.node_kind].append(bad_node)

        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attr_schema = base_node_schema.get_attribute("display_label")
        hfid_attr_schema = base_node_schema.get_attribute("human_friendly_id")

        processed = 0
        for kind, nodes in nodes_by_kind.items():
            try:
                schema = schema_branch.get(name=kind, duplicate=False)
            except SchemaNotFoundError:
                console.print(f"  Schema not found for kind {kind}, skipping {len(nodes)} nodes")
                processed += len(nodes)
                continue

            if isinstance(schema, GenericSchema):
                console.print(f"  Skipping GenericSchema kind {kind}")
                processed += len(nodes)
                continue

            for batch_start in range(0, len(nodes), self.update_batch_size):
                batch = nodes[batch_start : batch_start + self.update_batch_size]

                dl_values, hfid_values = await self._compute_values_for_batch(
                    db=db, load_branch=load_branch, schema=schema, batch=batch, errors=errors
                )

                if dl_values:
                    update_query = await UpdateAttributeValuesQuery.init(
                        db=db,
                        branch=write_branch,
                        attribute_schema=display_label_attr_schema,
                        values_by_id_map=dl_values,
                        at=at,
                    )
                    await update_query.execute(db=db)

                if hfid_values:
                    update_query = await UpdateAttributeValuesQuery.init(
                        db=db,
                        branch=write_branch,
                        attribute_schema=hfid_attr_schema,
                        values_by_id_map=hfid_values,
                        at=at,
                    )
                    await update_query.execute(db=db)

                processed += len(batch)
                if progress and progress_task is not None:
                    progress.update(progress_task, completed=processed)

        return MigrationResult(errors=errors)

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)

        result = MigrationResult()

        try:
            with Progress(console=console) as progress:
                # Fix bad values on default branch (branch-aware nodes)
                bad_nodes_default = await self._find_bad_nodes(db=db, value_branch_names=[default_branch_name])
                if bad_nodes_default:
                    task = progress.add_task(
                        f"Fixing {len(bad_nodes_default)} nodes with bad display_label/HFID on default branch",
                        total=len(bad_nodes_default),
                    )
                    default_result = await self._recompute_and_update(
                        db=db,
                        write_branch=default_branch,
                        load_branch=default_branch,
                        schema_branch=schema_branch,
                        bad_nodes=bad_nodes_default,
                        at=at,
                        progress=progress,
                        progress_task=task,
                    )
                    result.errors.extend(default_result.errors)
                else:
                    console.print("No nodes with bad values found on default branch")

                # Fix bad values on global branch (branch-agnostic nodes)
                global_branch = await Branch.get_by_name(db=db, name=GLOBAL_BRANCH_NAME)
                bad_nodes_global = await self._find_bad_nodes(db=db, value_branch_names=[GLOBAL_BRANCH_NAME])
                if bad_nodes_global:
                    task = progress.add_task(
                        f"Fixing {len(bad_nodes_global)} nodes with bad display_label/HFID on global branch",
                        total=len(bad_nodes_global),
                    )
                    global_result = await self._recompute_and_update(
                        db=db,
                        write_branch=global_branch,
                        load_branch=default_branch,
                        schema_branch=schema_branch,
                        bad_nodes=bad_nodes_global,
                        at=at,
                        progress=progress,
                        progress_task=task,
                    )
                    result.errors.extend(global_result.errors)
                else:
                    console.print("No nodes with bad values found on global branch")

        except Exception as exc:
            result.errors.append(str(exc))

        return result

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)

        # For user branches, bad HAS_VALUE edges are on this branch, but structural edges
        # (HAS_ATTRIBUTE, IS_PART_OF) may be on the default or global branch.
        root_node = await get_root_node(db=db, initialize=False)
        structural_branches = [branch.name, root_node.default_branch, GLOBAL_BRANCH_NAME]
        bad_nodes = await self._find_bad_nodes(
            db=db, value_branch_names=[branch.name], structural_branch_names=structural_branches
        )
        if not bad_nodes:
            console.print(f"No nodes with bad values found on branch {branch.name}")
            return MigrationResult()

        try:
            with Progress(console=console) as progress:
                task = progress.add_task(
                    f"Fixing {len(bad_nodes)} nodes with bad display_label/HFID on branch {branch.name}",
                    total=len(bad_nodes),
                )
                return await self._recompute_and_update(
                    db=db,
                    write_branch=branch,
                    load_branch=branch,
                    schema_branch=schema_branch,
                    bad_nodes=bad_nodes,
                    at=at,
                    progress=progress,
                    progress_task=task,
                )
        except Exception as exc:
            return MigrationResult(errors=[str(exc)])



from typing import Any

from infrahub.core.diff.model.path import NodeIdentifier
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffDropNodesQuery(Query):
    name = "enriched_diff_drop_nodes"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, enriched_diff_uuid: str, node_identifiers: list[NodeIdentifier], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enriched_diff_uuid = enriched_diff_uuid
        self.node_identifiers = node_identifiers

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_root_uuid": self.enriched_diff_uuid,
            "node_uuids": [ni.uuid for ni in self.node_identifiers],
            "node_identifiers_map": {ni.uuid: ni.kind for ni in self.node_identifiers},
        }
        query = """
        MATCH (d_root:DiffRoot {uuid: $diff_root_uuid})
        MATCH (d_root)-[:DIFF_HAS_NODE]->(dn:DiffNode)
        WHERE dn.uuid IN $node_uuids
        AND dn.kind IN $node_identifiers_map[dn.uuid]
        OPTIONAL MATCH (dn)-[:DIFF_HAS_ATTRIBUTE]-(da:DiffAttribute)
        OPTIONAL MATCH (da)-[*]->(diff_thing)
        DETACH DELETE diff_thing
        DETACH DELETE da
        WITH dn
        OPTIONAL MATCH (dn)-[:DIFF_HAS_RELATIONSHIP]->(dr:DiffRelationship)
        OPTIONAL MATCH (dr)-[:DIFF_HAS_ELEMENT]->(dre:DiffRelationshipElement)
        OPTIONAL MATCH (dre)-[*]->(diff_thing)
        DETACH DELETE diff_thing
        DETACH DELETE dre
        DETACH DELETE dr
        DETACH DELETE dn
        """
        self.add_to_query(query=query)



from pydantic import Field

from infrahub.message_bus import InfrahubMessage, InfrahubResponse, InfrahubResponseData

ROUTING_KEY = "git.file.get"


class GitFileGet(InfrahubMessage):
    """Read a file from a Git repository."""

    commit: str = Field(..., description="The commit id to use to access the file")
    file: str = Field(..., description="The path and filename within the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The kind of the repository")


class GitFileGetResponseData(InfrahubResponseData):
    content: str | None = None  # is empty if error_message / http_code are not empty
    error_message: str | None = None
    http_code: int | None = None


class GitFileGetResponse(InfrahubResponse):
    routing_key: str = ROUTING_KEY
    data: GitFileGetResponseData



from __future__ import annotations

from typing import TYPE_CHECKING

from graphql.language.ast import (
    DocumentNode,
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    FieldDefinitionNode,
    InputObjectTypeDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    NamedTypeNode,
    ObjectTypeDefinitionNode,
)

if TYPE_CHECKING:
    from graphql import DefinitionNode


def _sort_arguments(args: tuple[InputValueDefinitionNode, ...] | None) -> list[InputValueDefinitionNode] | None:
    """Sort arguments (filters) of a field alphabetically by name.

    Args:
        args: List of input value definition nodes to sort, or None.

    Returns:
        Sorted list of input value definition nodes, or None if input was None.
    """
    if not args:
        return None
    return sorted(args, key=lambda a: a.name.value)


def _sort_fields(fields: tuple[FieldDefinitionNode, ...] | None) -> list[FieldDefinitionNode] | None:
    """Sort fields and their arguments alphabetically.

    Args:
        fields: List of field definition nodes to sort, or None.

    Returns:
        Sorted list of field definition nodes with sorted arguments, or None if input was None.
    """
    if not fields:
        return None
    sorted_fields = []
    for field in sorted(fields, key=lambda fld: fld.name.value):
        sorted_args = _sort_arguments(field.arguments)
        sorted_fields.append(
            FieldDefinitionNode(
                name=field.name,
                type=field.type,
                arguments=sorted_args,
                directives=field.directives,
                description=field.description,
                loc=field.loc,
            )
        )
    return sorted_fields


def _sort_enum_values(values: tuple[EnumValueDefinitionNode, ...] | None) -> list[EnumValueDefinitionNode] | None:
    """Sort enum values alphabetically by name.

    Args:
        values: List of enum value definition nodes to sort, or None.

    Returns:
        Sorted list of enum value definition nodes, or None if input was None.
    """
    if not values:
        return None
    return sorted(values, key=lambda v: v.name.value)


def _sort_input_fields(fields: tuple[InputValueDefinitionNode, ...] | None) -> list[InputValueDefinitionNode] | None:
    """Sort input object fields alphabetically by name.

    Args:
        fields: List of input value definition nodes to sort, or None.

    Returns:
        Sorted list of input value definition nodes, or None if input was None.
    """
    if not fields:
        return None
    return sorted(fields, key=lambda f: f.name.value)


def _sort_interfaces(interfaces: tuple[NamedTypeNode, ...] | None) -> list[NamedTypeNode] | None:
    """Sort interface implementations alphabetically by name.

    Args:
        interfaces: Tuple of named type nodes representing interfaces, or None.

    Returns:
        Sorted list of named type nodes, or None if input was None.
    """
    if not interfaces:
        return None
    return sorted(interfaces, key=lambda i: i.name.value)


def sort_schema_ast(document: DocumentNode) -> DocumentNode:
    """Return a new DocumentNode with all definitions, fields, and arguments sorted alphabetically.

    This function recursively sorts all GraphQL schema elements including:
    - Type definitions (objects, interfaces, enums, input objects)
    - Field definitions and their arguments
    - Enum values
    - Input object fields

    Args:
        document: The GraphQL document node containing schema definitions.

    Returns:
        A new DocumentNode with all elements sorted alphabetically by name.
    """

    sorted_definitions: list[
        ObjectTypeDefinitionNode
        | InterfaceTypeDefinitionNode
        | EnumTypeDefinitionNode
        | InputObjectTypeDefinitionNode
        | DefinitionNode
    ] = []

    for definition in sorted(
        document.definitions, key=lambda d: getattr(d.name, "value", "") if hasattr(d, "name") and d.name else ""
    ):
        if isinstance(definition, (ObjectTypeDefinitionNode, InterfaceTypeDefinitionNode)):
            sorted_fields = _sort_fields(definition.fields)
            sorted_interfaces = _sort_interfaces(definition.interfaces)
            sorted_definitions.append(
                definition.__class__(
                    name=definition.name,
                    interfaces=sorted_interfaces,
                    directives=definition.directives,
                    fields=sorted_fields,
                    description=definition.description,
                    loc=definition.loc,
                )
            )

        elif isinstance(definition, EnumTypeDefinitionNode):
            sorted_values = _sort_enum_values(definition.values)
            sorted_definitions.append(
                EnumTypeDefinitionNode(
                    name=definition.name,
                    directives=definition.directives,
                    values=sorted_values,
                    description=definition.description,
                    loc=definition.loc,
                )
            )
        elif isinstance(definition, InputObjectTypeDefinitionNode):
            sorted_inputs = _sort_input_fields(definition.fields)
            sorted_definitions.append(
                InputObjectTypeDefinitionNode(
                    name=definition.name,
                    directives=definition.directives,
                    fields=sorted_inputs,
                    description=definition.description,
                    loc=definition.loc,
                )
            )
        else:
            sorted_definitions.append(definition)

    return DocumentNode(definitions=sorted_definitions)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub.core.constants import MetadataOptions

if TYPE_CHECKING:
    from infrahub.core.timestamp import Timestamp


@dataclass(frozen=True)
class MetadataQueryOptions:
    node_level: MetadataOptions = MetadataOptions.NONE
    attribute_level: MetadataOptions = MetadataOptions.NONE
    relationship_level: MetadataOptions = MetadataOptions.NONE

    def __or__(self, other: MetadataQueryOptions) -> MetadataQueryOptions:
        return MetadataQueryOptions(
            node_level=self.node_level | other.node_level,
            attribute_level=self.attribute_level | other.attribute_level,
            relationship_level=self.relationship_level | other.relationship_level,
        )


@dataclass
class MetadataInfo:
    created_at: Timestamp | None = None
    created_by: str | None = None
    updated_at: Timestamp | None = None
    updated_by: str | None = None



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from infrahub_sdk.template.exceptions import JinjaTemplateError
from rich.progress import Progress, TaskID

from infrahub.computed_attribute.jinja2 import InfrahubJinja2Template
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.query.update_attribute_values import UpdateAttributeValuesQuery
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)
from infrahub.core.query import Query, QueryType
from infrahub.core.schema.definitions.core.permission import core_global_permission, core_object_permission

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.schema import AttributeSchema
    from infrahub.database import InfrahubDatabase


console = get_migration_console()


@dataclass
class GlobalPermissionAttributes:
    """Attributes for a GlobalPermission node."""

    node_uuid: str
    action: str | None
    decision: int | None


@dataclass
class ObjectPermissionAttributes:
    """Attributes for an ObjectPermission node."""

    node_uuid: str
    namespace: str | None
    name: str | None
    action: str | None
    decision: int | None


class GetGlobalPermissionAttributesQuery(Query):
    """Get GlobalPermission nodes with their attribute values for computing display_label."""

    name = "get_global_permission_attributes"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {"branch_names": [GLOBAL_BRANCH_NAME]}

        query = """
MATCH (n:Node {kind: "CoreGlobalPermission"})-[e:IS_PART_OF]->(:Root)
WHERE e.branch IN $branch_names
AND e.status = "active"
AND e.to IS NULL
AND NOT exists((n)-[:IS_PART_OF {branch: e.branch, status: "deleted"}]->(:Root))
WITH DISTINCT n
// Get action attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(action_attr:Attribute {name: "action"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (action_attr)-[hv:HAS_VALUE]->(action_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN action_val.value AS action_value
}
// Get decision attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(decision_attr:Attribute {name: "decision"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (decision_attr)-[hv:HAS_VALUE]->(decision_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN decision_val.value AS decision_value
}
RETURN n.uuid AS node_uuid, action_value, decision_value
        """
        self.add_to_query(query)
        self.return_labels = ["node_uuid", "action_value", "decision_value"]

    def get_permission_results(self) -> list[GlobalPermissionAttributes]:
        """Return results as GlobalPermissionAttributes dataclasses."""
        return [
            GlobalPermissionAttributes(
                node_uuid=result.get_as_type("node_uuid", return_type=str),
                action=result.get_as_str("action_value"),
                decision=result.get_as_type("decision_value", return_type=int),
            )
            for result in self.get_results()
        ]


class GetObjectPermissionAttributesQuery(Query):
    """Get ObjectPermission nodes with their attribute values for computing display_label (default branch)."""

    name = "get_object_permission_attributes"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {"branch_names": [registry.default_branch]}

        query = """
MATCH (n:Node {kind: "CoreObjectPermission"})-[e:IS_PART_OF]->(:Root)
WHERE e.branch IN $branch_names
AND e.status = "active"
AND e.to IS NULL
AND NOT exists((n)-[:IS_PART_OF {branch: e.branch, status: "deleted"}]->(:Root))
WITH DISTINCT n
// Get namespace attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(ns_attr:Attribute {name: "namespace"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (ns_attr)-[hv:HAS_VALUE]->(ns_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN ns_val.value AS namespace_value
}
// Get name attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(name_attr:Attribute {name: "name"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (name_attr)-[hv:HAS_VALUE]->(name_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN name_val.value AS name_value
}
// Get action attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(action_attr:Attribute {name: "action"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (action_attr)-[hv:HAS_VALUE]->(action_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN action_val.value AS action_value
}
// Get decision attribute value
CALL (n) {
    MATCH (n)-[ha:HAS_ATTRIBUTE]->(decision_attr:Attribute {name: "decision"})
    WHERE ha.branch IN $branch_names AND ha.status = "active" AND ha.to IS NULL
    MATCH (decision_attr)-[hv:HAS_VALUE]->(decision_val:AttributeValue)
    WHERE hv.branch IN $branch_names AND hv.status = "active" AND hv.to IS NULL
    RETURN decision_val.value AS decision_value
}
RETURN n.uuid AS node_uuid, namespace_value, name_value, action_value, decision_value
        """
        self.add_to_query(query)
        self.return_labels = ["node_uuid", "namespace_value", "name_value", "action_value", "decision_value"]

    def get_permission_results(self) -> list[ObjectPermissionAttributes]:
        """Return results as ObjectPermissionAttributes dataclasses."""
        return [
            ObjectPermissionAttributes(
                node_uuid=result.get_as_type("node_uuid", return_type=str),
                namespace=result.get_as_str("namespace_value"),
                name=result.get_as_str("name_value"),
                action=result.get_as_str("action_value"),
                decision=result.get_as_type("decision_value", return_type=int),
            )
            for result in self.get_results()
        ]


class GetObjectPermissionAttributesBranchQuery(Query):
    """Get ObjectPermission nodes with their attribute values on a non-default branch."""

    name = "get_object_permission_attributes_branch"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter_r, branch_filter_params_r = self.branch.get_query_filter_path(
            at=self.at, variable_name="r", params_prefix="r_"
        )
        branch_filter_r2, branch_filter_params_r2 = self.branch.get_query_filter_path(
            at=self.at, variable_name="r2", params_prefix="r2_"
        )
        self.params = {**branch_filter_params_r, **branch_filter_params_r2}

        query = """
MATCH (n:Node {kind: "CoreObjectPermission"})
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter_r)s
    RETURN r AS is_part_of_e
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_part_of_e
WHERE is_part_of_e.status = "active"
// Get namespace attribute value
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(ns_attr:Attribute {name: "namespace"})
    WHERE %(branch_filter_r)s
    WITH ns_attr, r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH ns_attr, r
    WHERE r.status = "active"
    MATCH (ns_attr)-[r2:HAS_VALUE]->(ns_val:AttributeValue)
    WHERE %(branch_filter_r2)s
    WITH ns_val, r2
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC
    LIMIT 1
    WITH ns_val, r2
    WHERE r2.status = "active"
    RETURN ns_val.value AS namespace_value
}
// Get name attribute value
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(name_attr:Attribute {name: "name"})
    WHERE %(branch_filter_r)s
    WITH name_attr, r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH name_attr, r
    WHERE r.status = "active"
    MATCH (name_attr)-[r2:HAS_VALUE]->(name_val:AttributeValue)
    WHERE %(branch_filter_r2)s
    WITH name_val, r2
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC
    LIMIT 1
    WITH name_val, r2
    WHERE r2.status = "active"
    RETURN name_val.value AS name_value
}
// Get action attribute value
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(action_attr:Attribute {name: "action"})
    WHERE %(branch_filter_r)s
    WITH action_attr, r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH action_attr, r
    WHERE r.status = "active"
    MATCH (action_attr)-[r2:HAS_VALUE]->(action_val:AttributeValue)
    WHERE %(branch_filter_r2)s
    WITH action_val, r2
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC
    LIMIT 1
    WITH action_val, r2
    WHERE r2.status = "active"
    RETURN action_val.value AS action_value
}
// Get decision attribute value
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(decision_attr:Attribute {name: "decision"})
    WHERE %(branch_filter_r)s
    WITH decision_attr, r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH decision_attr, r
    WHERE r.status = "active"
    MATCH (decision_attr)-[r2:HAS_VALUE]->(decision_val:AttributeValue)
    WHERE %(branch_filter_r2)s
    WITH decision_val, r2
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC
    LIMIT 1
    WITH decision_val, r2
    WHERE r2.status = "active"
    RETURN decision_val.value AS decision_value
}
RETURN n.uuid AS node_uuid, namespace_value, name_value, action_value, decision_value
        """ % {"branch_filter_r": branch_filter_r, "branch_filter_r2": branch_filter_r2}

        self.add_to_query(query)
        self.return_labels = ["node_uuid", "namespace_value", "name_value", "action_value", "decision_value"]

    def get_permission_results(self) -> list[ObjectPermissionAttributes]:
        """Return results as ObjectPermissionAttributes dataclasses."""
        return [
            ObjectPermissionAttributes(
                node_uuid=result.get_as_type("node_uuid", return_type=str),
                namespace=result.get_as_str("namespace_value"),
                name=result.get_as_str("name_value"),
                action=result.get_as_str("action_value"),
                decision=result.get_as_type("decision_value", return_type=int),
            )
            for result in self.get_results()
        ]


class Migration062(MigrationRequiringRebase):
    """Recompute display_label for all permission nodes (CoreObjectPermission and CoreGlobalPermission)."""

    name: str = "062_recompute_permission_display_labels"
    minimum_version: int = 61
    update_batch_size: int = 1000

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def _compute_object_permission_display_labels(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        attribute_schema: AttributeSchema,
        progress: Progress | None = None,
        update_task: TaskID | None = None,
    ) -> None:
        """Compute and update display_label for ObjectPermission nodes."""
        if not core_object_permission.display_label:
            return

        query: GetObjectPermissionAttributesQuery | GetObjectPermissionAttributesBranchQuery
        if branch.is_default:
            query = await GetObjectPermissionAttributesQuery.init(db=db)
        else:
            query = await GetObjectPermissionAttributesBranchQuery.init(db=db, branch=branch)
        await query.execute(db=db)
        results = query.get_permission_results()

        if not results:
            return

        values_by_id: dict[str, str] = {}
        jinja_template = InfrahubJinja2Template(template=core_object_permission.display_label)

        for row in results:
            if row.namespace is None or row.name is None or row.action is None or row.decision is None:
                continue

            variables = {
                "namespace__value": row.namespace,
                "name__value": row.name,
                "action__value": row.action,
                "decision__value": row.decision,
            }

            try:
                rendered = await jinja_template.render(variables=variables)
                if rendered is not None:
                    values_by_id[row.node_uuid] = rendered
            except JinjaTemplateError as exc:
                console.log(f"[yellow]Warning: Failed to render display_label for node {row.node_uuid}: {exc}[/yellow]")
                continue

        if values_by_id:
            for offset in range(0, len(values_by_id), self.update_batch_size):
                batch_ids = list(values_by_id.keys())[offset : offset + self.update_batch_size]
                batch_map = {k: values_by_id[k] for k in batch_ids}

                update_query = await UpdateAttributeValuesQuery.init(
                    db=db, branch=branch, attribute_schema=attribute_schema, values_by_id_map=batch_map
                )
                await update_query.execute(db=db)

                if progress is not None and update_task is not None:
                    progress.update(update_task, advance=len(batch_ids))

    async def _compute_global_permission_display_labels(
        self,
        db: InfrahubDatabase,
        attribute_schema: AttributeSchema,
        progress: Progress | None = None,
        update_task: TaskID | None = None,
    ) -> None:
        """Compute and update display_label for GlobalPermission nodes (branch-agnostic)."""
        if not core_global_permission.display_label:
            return

        query = await GetGlobalPermissionAttributesQuery.init(db=db, branch_agnostic=True)
        await query.execute(db=db)
        results = query.get_permission_results()

        if not results:
            return

        values_by_id: dict[str, str] = {}
        jinja_template = InfrahubJinja2Template(template=core_global_permission.display_label)

        for row in results:
            if row.action is None or row.decision is None:
                continue

            variables = {"action__value": row.action, "decision__value": row.decision}

            try:
                rendered = await jinja_template.render(variables=variables)
                if rendered is not None:
                    values_by_id[row.node_uuid] = rendered
            except JinjaTemplateError as exc:
                console.log(f"[yellow]Warning: Failed to render display_label for node {row.node_uuid}: {exc}[/yellow]")
                continue

        if values_by_id:
            global_branch = await Branch.get_by_name(db=db, name=GLOBAL_BRANCH_NAME)
            for offset in range(0, len(values_by_id), self.update_batch_size):
                batch_ids = list(values_by_id.keys())[offset : offset + self.update_batch_size]
                batch_map = {k: values_by_id[k] for k in batch_ids}

                update_query = await UpdateAttributeValuesQuery.init(
                    db=db, branch=global_branch, attribute_schema=attribute_schema, values_by_id_map=batch_map
                )
                await update_query.execute(db=db)

                if progress is not None and update_task is not None:
                    progress.update(update_task, advance=len(batch_ids))

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db

        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)

        schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")

        obj_count_query = await GetObjectPermissionAttributesQuery.init(db=db)
        await obj_count_query.execute(db=db)
        obj_permission_count = len(obj_count_query.get_permission_results())

        global_count_query = await GetGlobalPermissionAttributesQuery.init(db=db, branch_agnostic=True)
        await global_count_query.execute(db=db)
        global_permission_count = len(global_count_query.get_permission_results())

        total_count = obj_permission_count + global_permission_count

        try:
            with Progress(console=console) as progress:
                if total_count > 0:
                    update_task = progress.add_task(
                        f"Recomputing display_label for {total_count} permissions", total=total_count
                    )

                    await self._compute_object_permission_display_labels(
                        db=db,
                        branch=default_branch,
                        attribute_schema=display_label_attribute_schema,
                        progress=progress,
                        update_task=update_task,
                    )
                    await self._compute_global_permission_display_labels(
                        db=db,
                        attribute_schema=display_label_attribute_schema,
                        progress=progress,
                        update_task=update_task,
                    )

        except Exception as exc:
            error_msg = str(exc) or f"{type(exc).__name__}: {repr(exc)}"
            return MigrationResult(errors=[error_msg])

        return MigrationResult()

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        """Execute migration on non-default branches (only ObjectPermission, GlobalPermission is branch-agnostic)."""
        db = migration_input.db

        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)
        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")

        try:
            await self._compute_object_permission_display_labels(
                db=db, branch=branch, attribute_schema=display_label_attribute_schema
            )
        except Exception as exc:
            error_msg = str(exc) or f"{type(exc).__name__}: {repr(exc)}"
            return MigrationResult(errors=[error_msg])

        return MigrationResult()
